from __future__ import division
# vim: tabstop=2 shiftwidth=2 softtabstop=2 expandtab autoindent smarttab
from builtins import str
from builtins import map
from builtins import range
from past.utils import old_div
from builtins import object
import abc
import collections
import copy
import hashlib
import json
import logging
import math
import os
import pickle
import random
from fn import _
import argparse
from datetime import datetime
import numpy
import inspect
import sys
from future.utils import with_metaclass
from functools import reduce

log = logging.getLogger(__name__)
argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--list-params', '-lp',
                       help='list available parameter classes')


class ConfigurationManipulatorBase(with_metaclass(abc.ABCMeta, object)):
  """
  abstract interface for objects used by search techniques to mutate
  configurations
  """

  # List of file formats, which can be extended by subclasses. Used in
  # write_to_file() and load_from_file().  Objects in list must define
  # load(fd) and dump(cfg, fd).
  FILE_FORMATS = {'default': json, 'json': json,
                  'pickle': pickle, 'pk': pickle}

  def validate(self, config):
    """is the given config valid???"""
    return all(map(_.validate(config), self.parameters(config)))

  def normalize(self, config):
    """mutate config into canonical form"""
    for param in self.parameters(config):
      param.normalize(config)

  def set_search_driver(self, search_driver):
    """called exactly once during setup"""
    pass

  def copy(self, config):
    """produce copy of config"""
    return copy.deepcopy(config)

  def parameters_dict(self, config):
    """convert self.parameters() to a dictionary by name"""
    return dict([(p.name, p) for p in self.parameters(config)])

  def param_names(self, *args):
    """return union of parameter names in args"""
    return sorted(reduce(set.union,
                         [set(map(_.name, self.parameters(cfg)))
                          for cfg in args]))

  def linear_config(self, a, cfg_a, b, cfg_b, c, cfg_c):
    """return a configuration that is a linear combination of 3 other configs"""
    dst = self.copy(cfg_a)
    dst_params = self.proxy(dst)
    for k in self.param_names(dst, cfg_a, cfg_b, cfg_c):
      dst_params[k].op4_set_linear(cfg_a, cfg_b, cfg_c, a, b, c)
    return dst

  def _get_serializer(self, filename, format=None):
    """
    Extract the correct file format serializer from self.FILE_FORMATS.
    Guess the format by extension if one is not given.
    """
    if format is None:
      format = os.path.splitext(filename)[1].lower().replace('.', '')
    if format not in self.FILE_FORMATS:
      serializer = self.FILE_FORMATS['default']
      if len(self.FILE_FORMATS) > 1:
        log.warning('Unknown file format "%s", using "%s" instead', format,
                    serializer.__name__)
    else:
      serializer = self.FILE_FORMATS[format]
    return serializer

  def save_to_file(self, cfg, filename, format=None):
    """
    Write cfg to filename.  Guess the format by extension if one is not given.
    """
    with open(filename, 'wb') as fd:
      self._get_serializer(filename, format).dump(cfg, fd)

  def load_from_file(self, filename, format=None):
    """
    Read cfg from filename.  Guess the format by extension if one is not given.
    """
    with open(filename, 'rb') as fd:
      return self._get_serializer(filename, format).load(fd)

  def proxy(self, cfg):
    return ManipulatorProxy(self, cfg)

  @abc.abstractmethod
  def random(self):
    """produce a random initial configuration"""
    return

  @abc.abstractmethod
  def parameters(self, config):
    """return a list of of Parameter objects"""
    return list()

  @abc.abstractmethod
  def hash_config(self, config):
    """produce unique hash value for the given config"""
    return


class ConfigurationManipulator(ConfigurationManipulatorBase):
  """
  a configuration manipulator using a fixed set of parameters and storing
  configs in a dict-like object
  """

  def __init__(self, params=None, config_type=dict, seed_config=None, **kwargs):
    if params is None:
      params = []
    self.params = list(params)
    self.config_type = config_type
    self.search_driver = None
    self._seed_config = seed_config
    super(ConfigurationManipulator, self).__init__(**kwargs)
    for p in self.params:
      p.parent = self

  def add_parameter(self, p):
    p.set_parent(self)
    self.params.append(p)

    #TODO sub parameters should be recursed on
    # not currently an issue since no doubly-nested sub-parameters
    sub_params = p.sub_parameters()
    for sp in sub_params:
      sp.set_parent(p)
    self.params.extend(sub_params)

  def set_search_driver(self, search_driver):
    self.search_driver = search_driver

  def seed_config(self):
    """produce a fixed seed configuration"""
    if self._seed_config:
      cfg = copy.deepcopy(self._seed_config)
    else:
      cfg = self.config_type()
      for p in self.params:
        if not isinstance(p.name, str) or '/' not in p.name:
          cfg[p.name] = p.seed_value()
    return cfg

  def random(self):
    """produce a random configuration"""
    cfg = self.seed_config()
    for p in self.parameters(cfg):
      p.op1_randomize(cfg)
    return cfg

  def parameters(self, config):
    """return a list of Parameter objects"""
    if type(config) is not self.config_type:
      log.error("wrong type, expected %s got %s",
                str(self.config_type),
                str(type(config)))
      raise TypeError()
    return self.params

  def parameters_to_json(self):
    """
    output information about the parameters in this manipulator in json format:
    [ConfigurationManipulator,{pinfo:count,pinfo:count ...}]
    where pinfo has a similar form to describe the parameter's sub-parameters:
    [param_name,{pinfo:count,pinfo:count ...}]
    """
    def param_info_to_json(param, sub_parameters):
      """
      recursively output information about a parameter and its subparameters in a json format:

      [parameter_name, {subparam_info:count,subparam_info:count,...}]
      or if no subparams
      [parameter_name,{}]

      where subparam_info are sorted alphabetically. Note we can't directly use json since
      sets/dictionaries aren't always ordered by key
      """
      sub_parameter_counts = {}
      # build the string
      if isinstance(param, str):
        param_name = param
      else:
        param_name = param.__class__.__name__
      out = ['[', param_name, ',{']

      if len(sub_parameters) > 0:
        # count sub params
        for sp in sub_parameters:
          spout = param_info_to_json(sp, sp.sub_parameters())
          sub_parameter_counts[spout] = sub_parameter_counts.get(spout, 0) + 1
        # add the count map in sorted order
        for sp in sorted(sub_parameter_counts):
          out.append(sp)
          out.append(':')
          out.append(str(sub_parameter_counts[sp]))
          out.append(',')
        out.pop() # remove trailing comma

      out.append('}]')
      return ''.join(out)

    # filter out subparameters to avoid double counting
    params = [p for p in self.params if p.parent is self]
    return param_info_to_json(self, params)

  def hash_config(self, config):
    """produce unique hash value for the given config"""
    m = hashlib.sha256()
    params = list(self.parameters(config))
    params.sort(key=_.name)
    for i, p in enumerate(params):
      m.update(str(p.name).encode())
      m.update(p.hash_value(config))
      m.update(str(i).encode())
      m.update(b"|")
    return m.hexdigest()

  def search_space_size(self):
    """estimate the size of the search space, not precise"""
    return reduce(_ * _, [x.search_space_size() for x in self.params])

  def difference(self, cfg1, cfg2):
    cfg = self.copy(cfg1)
    for param in self.parameters(cfg1):
      if param.is_primitive(cfg1):
        # TODO: check range
        param.set_value(cfg, param.get_value(cfg1) - param.get_value(cfg2))
      else:
        pass
    return cfg

  def applySVs(self, cfg, sv_map, args, kwargs):
    """
    Apply operators to each parameter according to given map. Updates cfg.
    Parameters with no operators specified are not updated.
    cfg: configuration data
    sv_map: python dict that maps string parameter name to class method name
    arg_map: python dict that maps string parameter name to class method
    arguments
    """
    # TODO: check consistency between sv_map and cfg
    param_dict = self.parameters_dict(cfg)
    for pname in self.param_names(cfg):
      param = param_dict[pname]
      getattr(param, sv_map[pname])(cfg, *args[pname], **kwargs[pname])


class Parameter(with_metaclass(abc.ABCMeta, object)):
  """
  abstract base class for parameters in a ConfigurationManipulator
  """

  def __init__(self, name):
    self.name = name
    self.parent = None
    super(Parameter, self).__init__()

  def _to_storage_type(self, val):
    """hook to support transformation applied while stored"""
    return val

  def _from_storage_type(self, sval):
    """hook to support transformation applied while stored"""
    return sval

  def _read_node(self, config):
    """hook to support different storage structures"""
    node = config
    if not isinstance(self.name, str):
      return node, self.name
    name_parts = self.name.split('/')
    for part in name_parts[:-1]:
      if isinstance(node, list):
        part = int(part)
      node = node[part]
    part = name_parts[-1]
    if isinstance(node, list):
      part = int(part)
    return node, part

  def _get(self, config):
    """hook to support different storage structures"""
    node, part = self._read_node(config)
    return self._from_storage_type(node[part])

  def _set(self, config, v):
    """hook to support different storage structures"""
    node, part = self._read_node(config)
    node[part] = self._to_storage_type(v)

  def set_parent(self, manipulator):
    self.parent = manipulator

  def validate(self, config):
    """is the given config valid???"""
    return True

  def is_primitive(self, ignored=None):
    return isinstance(self, PrimitiveParameter)

  def is_permutation(self, ignored=None):
    return isinstance(self, PermutationParameter)

  def manipulators(self, config):
    """
    a list of manipulator functions to change this value in the config
    manipulators must be functions that take a config and change it in place

    default implementation just has op1_randomize as only operation
    """
    return [self.op1_randomize]

  def normalize(self, config):
    """
    mutate this parameter into a canonical form
    """
    pass

  def sub_parameters(self):
    """
    additional parameters added with this parameter
    """
    return []

  @abc.abstractmethod
  def op1_randomize(self, cfg):
    """
    Set this parameter's value in a configuration to a random value

    :param config: the configuration to be changed
    """
    pass

  @abc.abstractmethod
  def seed_value(self):
    """some legal value of this parameter (for creating initial configs)"""
    return

  @abc.abstractmethod
  def copy_value(self, src, dst):
    """copy the value of this parameter from src to dst config"""
    pass

  @abc.abstractmethod
  def same_value(self, cfg1, cfg2):
    """test if cfg1 and cfg2 have the same value of this parameter"""
    return

  @abc.abstractmethod
  def hash_value(self, config):
    """produce unique hash for this value in the config"""
    return

  @abc.abstractmethod
  def op4_set_linear(self, cfg, cfg_a, cfg_b, cfg_c, a, b, c):
    """
    Sets the parameter value in a configuration to a linear combination of 3
    other configurations: :math:`a*cfg_a + b*cfg_b + c*cfg_c`

    :param cfg: the configuration to be changed
    :param cfg_a: a parent configuration
    :param cfg_b: a parent configuration
    :param cfg_c: a parent configuration
    :param a: weight for cfg_a
    :param b: weight for cfg_b
    :param c: weight for cfg_c
    """
    pass

  def search_space_size(self):
    return 1

  def op1_nop(self, cfg):
    """
    The 'null' operator. Does nothing.

    :param cfg: the configuration to be changed
    """
    pass

  # Stochastic variators
  def op3_swarm(self, cfg, cfg1, cfg2, c, c1, c2, *args, **kwargs):
    """
    Stochastically 'move' the parameter value in a configuration towards those
    in two parent configurations. This is done by calling :py:meth:`opn_stochastic_mix`

    :param cfg: the configuration to be changed
    :param cfg1: a parent configuration
    :param cfg2: a parent configuration
    :param c: weight of original configuration
    :param c1: weight for cfg1
    :param c2: weight for cfg2
    """
    # default to probabilistic treatment
    self.opn_stochastic_mix(cfg, [cfg, cfg1, cfg2], [c, c1, c2])

  def opn_stochastic_mix(self, cfg, cfgs, ratio, *args, **kwargs):
    """
    Stochastically recombine a list of parent values into a single result.

    This randomly copies a value from a list of parents configurations according
    to a list of weights.

    :param cfg: the configuration to be changed
    :param cfgs: a list of parent configurations
    :param ratio: a list of floats representing the weight of each configuration
     in cfgs

    """
    assert len(cfgs) == len(ratio)
    r = random.random()
    c = old_div(numpy.array(ratio, dtype=float), sum(ratio))
    for i in range(len(c)):
      if r < sum(c[:i + 1]):
        self.copy_value(cfg, cfgs[i])
        break


class PrimitiveParameter(with_metaclass(abc.ABCMeta, Parameter)):
  """
  An abstract interface implemented by parameters that represent a single
  dimension in a cartesian space in a legal range
  """

  def __init__(self, name, value_type=float, **kwargs):
    self.value_type = value_type
    super(PrimitiveParameter, self).__init__(name, **kwargs)

  def hash_value(self, config):
    """produce unique hash for this value in the config"""
    self.normalize(config)
    return hashlib.sha256(repr(self.get_value(config)).encode('utf-8')).hexdigest().encode('utf-8')

  def copy_value(self, src, dst):
    """copy the value of this parameter from src to dst config"""
    self.set_value(dst, self.get_value(src))

  def same_value(self, cfg1, cfg2):
    """test if cfg1 and cfg2 have the same value of this parameter"""
    return self.get_value(cfg1) == self.get_value(cfg2)

  def is_integer_type(self):
    """true if self.value_type can only represent integers"""
    return self.value_type(0) == self.value_type(0.1)

  def get_unit_value(self, config):
    """get_value scaled such that range is between 0.0 and 1.0"""
    low, high = self.legal_range(config)
    if self.is_integer_type():
      # account for rounding
      low -= 0.4999
      high += 0.4999
    val = self.get_value(config)
    if low < high:
      return old_div(float(val - low), float(high - low))
    else:
      if low > high:
        log.warning('invalid range for parameter %s, %s to %s',
                    self.name, low, high)
      # only a single legal value!
      return 0.0

  def set_unit_value(self, config, unit_value):
    """set_value scaled such that range is between 0.0 and 1.0"""
    assert 0.0 <= unit_value <= 1.0
    low, high = self.legal_range(config)
    if self.is_integer_type():
      # account for rounding
      low -= 0.4999
      high += 0.4999
    if low < high:
      val = unit_value * float(high - low) + low
      if self.is_integer_type():
        val = round(val)
      val = max(low, min(val, high))
      self.set_value(config, self.value_type(val))

  def op1_normal_mutation(self, cfg, sigma=0.1, *args, **kwargs):
    """
    apply normally distributed noise to this parameter's value in a
    configuration

    :param cfg: The configuration to be changed
    :param sigma: the std. deviation of the normally distributed noise on a unit
     scale
    """
    v = self.get_unit_value(cfg)
    v += random.normalvariate(0.0, sigma)
    # handle boundary cases by reflecting off the edge
    if v < 0.0:
      v *= -1.0
    if v > 1.0:
      v = 1.0 - (v % 1)
    self.set_unit_value(cfg, v)

  def op4_set_linear(self, cfg, cfg_a, cfg_b, cfg_c, a, b, c):
    """
    set the parameter value in a configuration to a linear combination of 3
    other configurations: :math:`a*cfg_a + b*cfg_b + c*cfg_c`

    :param cfg: The configuration to be changed
    :param cfg_a: a parent configuration
    :param cfg_b: a parent configuration
    :param cfg_c: a parent configuration
    :param a: weight for cfg_a
    :param b: weight for cfg_b
    :param c: weight for cfg_c
    """
    va = self.get_unit_value(cfg_a)
    vb = self.get_unit_value(cfg_b)
    vc = self.get_unit_value(cfg_c)
    v = a * va + b * vb + c * vc
    v = max(0.0, min(v, 1.0))

    self.set_unit_value(cfg, v)

  def manipulators(self, config):
    """
    a list of manipulator functions to change this value in the config
    manipulators must be functions that take a config and change it in place

    for primitive params default implementation is uniform random and normal
    """
    return [self.op1_randomize, self.op1_normal_mutation]

  @abc.abstractmethod
  def set_value(self, config, value):
    """assign this value in the given configuration"""
    pass

  @abc.abstractmethod
  def get_value(self, config):
    """retrieve this value from the given configuration"""
    return 0

  @abc.abstractmethod
  def legal_range(self, config):
    """return the legal range for this parameter, inclusive"""
    return 0, 1


class NumericParameter(PrimitiveParameter):
  """
  A parameter representing a number with a minimum and maximum value
  """
  def __init__(self, name, min_value, max_value, **kwargs):
    """min/max are inclusive"""
    assert min_value <= max_value
    super(NumericParameter, self).__init__(name, **kwargs)
    # after super call so self.value_type is initialized
    self.min_value = self.value_type(min_value)
    self.max_value = self.value_type(max_value)

  def seed_value(self):
    """some legal value of this parameter (for creating initial configs)"""
    return self.min_value

  def set_value(self, config, value):
    assert value >= self.min_value
    assert value <= self.max_value
    self._set(config, value)

  def get_value(self, config):
    return self._get(config)

  def legal_range(self, config):
    return self.min_value, self.max_value

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration to a random value in its legal
     range

    :param config: the configuration to be changed
    """
    if self.is_integer_type():
      self.set_value(config, random.randint(*self.legal_range(config)))
    else:
      self.set_value(config, random.uniform(*self.legal_range(config)))

  def op1_scale(self, cfg, k):
    """
    Scale this parameter's value in a configuration by a constant factor

    :param cfg: the configuration to be changed
    :param k: the constant factor to scale the parameter value by
    """
    v = self.get_value(cfg) * k
    v = max(self.min_value, min(self.max_value, v))
    self.set_value(cfg, v)

  def op3_difference(self, cfg, cfg1, cfg2):
    """
    Set this parameter's value in a configuration to the difference between this
    parameter's values in 2 other configs (cfg2 - cfg1)

    :param cfg: the configuration to be changed
    :param cfg1: The configuration whose parameter value is being subtracted
    :param cfg2: The configuration whose parameter value is subtracted from
    """
    v = self.get_value(cfg2) - self.get_value(cfg1)
    v = max(self.min_value, min(self.max_value, v))
    self.set_value(cfg, v)

  def opn_sum(self, cfg, *cfgs):
    """
    Set this parameter's value in a configuration to the sum of it's values in a
     list of configurations

    :param cfg: the configuration to be changed
    :param cfgs: a list of configurations to sum
    """
    v = sum([self.get_value(c) for c in cfgs])
    v = max(self.min_value, min(self.max_value, v))
    self.set_value(cfg, v)

  def search_space_size(self):
    if self.value_type is float:
      return 2 ** 32
    else:
      return self.max_value - self.min_value + 1  # inclusive range


class IntegerParameter(NumericParameter):
  """
  A parameter representing an integer value in a legal range
  """
  def __init__(self, name, min_value, max_value, **kwargs):
    """min/max are inclusive"""
    kwargs['value_type'] = int
    super(IntegerParameter, self).__init__(name, min_value, max_value, **kwargs)

  def op3_swarm(self, cfg, cfg1, cfg2, c=1, c1=0.5,
                c2=0.5, velocity=0, sigma=0.2, *args, **kwargs):
    """
    Simulates a single update step in particle swarm optimization by updating
    the current position and returning a new velocity.

    The new velocity is given by

    .. math:: c*velocity + r1*c1*(cfg1-cfg) + r2*c2*(cfg2-cfg)

    where r1 and r2 are random values between 0 and 1.

    The new current position is the new velocity with gaussian noise added.

    :param cfg: the configuration to be changed. Represents the current position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param c: the weight of the current velocity
    :param c1: weight of cfg1
    :param c2: weight of cfg2
    :param velocity: the old velocity
    :param sigma: standard deviation of the gaussian noise, on a unit-scale
    :return: the new velocity, a float

    """
    vmin, vmax = self.legal_range(cfg)
    k = vmax - vmin
    # calculate the new velocity
    v = velocity * c + (self.get_value(cfg1) - self.get_value(
        cfg)) * c1 * random.random() + (self.get_value(
        cfg2) - self.get_value(cfg)) * c2 * random.random()
    # Map velocity to continuous space with sigmoid
    s = old_div(k, (1 + numpy.exp(-v))) + vmin
    # Add Gaussian noise
    p = random.gauss(s, sigma * k)
    # Discretize and bound
    p = int(min(vmax, max(round(p), vmin)))
    self.set_value(cfg, p)
    return v


class FloatParameter(NumericParameter):
  def __init__(self, name, min_value, max_value, **kwargs):
    """min/max are inclusive"""
    kwargs['value_type'] = float
    super(FloatParameter, self).__init__(name, min_value, max_value, **kwargs)

  def op3_swarm(self, cfg, cfg1, cfg2, c=1, c1=0.5,
                c2=0.5, velocity=0, *args, **kwargs):
    """

    Simulates a single update step in particle swarm optimization by updating
    the current position and returning a new velocity.

    The new velocity is given by

    .. math:: c*velocity + r1*c1*(cfg1-cfg) + r2*c2*(cfg2-cfg)

    where r1 and r2 are random values between 0 and 1

    The new current position is the old current position offset by the new
    velocity:

    :param cfg: the configuration to be changed. Represents the current position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param c: the weight of the current velocity
    :param c1: weight of cfg1
    :param c2: weight of cfg2
    :param velocity: the old velocity
    :return: the new velocity, a float

    """
    vmin, vmax = self.legal_range(cfg)
    v = velocity * c + (self.get_value(cfg1) - self.get_value(
        cfg)) * c1 * random.random() + (self.get_value(
        cfg2) - self.get_value(cfg)) * c2 * random.random()
    p = self.get_value(cfg) + v
    p = min(vmax, max(p, vmin))
    self.set_value(cfg, p)
    return v


class ScaledNumericParameter(NumericParameter):
  """
  A Parameter that is stored in configurations normally, but has a scaled
  value when accessed using 'get_value'.
  Because search techniques interact with Parameters through get_value, these
  parameters are searched on a different scale (e.g. log scale).
  """

  @abc.abstractmethod
  def _scale(self, v):
    """
    called on a value when getting it from it's configuration. Transforms the
    actual value to the scale it is searched on
    """
    return v

  @abc.abstractmethod
  def _unscale(self, v):
    """
    called on a value when storing it. Transforms a value from it's search scale
    to it's actual value
    """
    return v

  def set_value(self, config, value):
    NumericParameter.set_value(self, config, self._unscale(value))

  def get_value(self, config):
    return self._scale(NumericParameter.get_value(self, config))

  def legal_range(self, config):
    return list(map(self._scale, NumericParameter.legal_range(self, config)))


class LogIntegerParameter(ScaledNumericParameter, FloatParameter):
  """
  an integer value that is searched on a log scale, but stored without scaling
  """

  def _scale(self, v):
    return math.log(v + 1.0 - self.min_value, 2.0)

  def _unscale(self, v):
    v = 2.0 ** v - 1.0 + self.min_value
    v = int(round(v))
    return v

  def legal_range(self, config):
    low, high = NumericParameter.legal_range(self, config)
    # increase the bounds account for rounding
    return self._scale(low - 0.4999), self._scale(high + 0.4999)


class LogFloatParameter(ScaledNumericParameter, FloatParameter):
  """
  a float parameter that is searched on a log scale, but stored without scaling
  """

  def _scale(self, v):
    return math.log(v + 1.0 - self.min_value, 2.0)

  def _unscale(self, v):
    v = 2.0 ** v - 1.0 + self.min_value
    return v


class PowerOfTwoParameter(ScaledNumericParameter, IntegerParameter):
  """
  An integer power of two, with a min and max value. Searched by the exponent
  """

  def __init__(self, name, min_value, max_value, **kwargs):
    kwargs['value_type'] = int
    assert min_value >= 1
    assert math.log(min_value, 2) % 1 == 0  # must be power of 2
    assert math.log(max_value, 2) % 1 == 0  # must be power of 2
    super(PowerOfTwoParameter, self).__init__(name, min_value, max_value,
                                              **kwargs)

  def _scale(self, v):
    return int(math.log(v, 2))

  def _unscale(self, v):
    return 2 ** int(v)

  def legal_range(self, config):
    return int(math.log(self.min_value, 2)), int(math.log(self.max_value, 2))

  def search_space_size(self):
    return int(math.log(super(PowerOfTwoParameter, self).search_space_size(), 2))


##################

class ComplexParameter(Parameter):
  """
  A non-cartesian parameter that can't be manipulated directly, but has a set
  of user defined manipulation functions
  """

  def copy_value(self, src, dst):
    """copy the value of this parameter from src to dst config"""
    self._set(dst, copy.deepcopy(self._get(src)))

  def same_value(self, cfg1, cfg2):
    """test if cfg1 and cfg2 have the same value of this parameter"""
    return self._get(cfg1) == self._get(cfg2)

  def hash_value(self, config):
    """produce unique hash for this value in the config"""
    self.normalize(config)
    return hashlib.sha256(repr(self._get(config))).hexdigest()

  def get_value(self, config):
    return self._get(config)

  def set_value(self, config, value):
    self._set(config, value)

  def op4_set_linear(self, cfg, cfg_a, cfg_b, cfg_c, a, b, c):
    """
    set this value to :math:`a*cfg_a + b*cfg_b + c*cfg_c`

    this operation is not possible in general with complex parameters but
    we make an attempt to "fake" it for common use cases

    basically a call to randomize unless after normalization,
    a = 1.0, b == -c, and cfg_b == cfg_c, in which case nothing is done

    :param cfg: the configuration to be changed
    :param cfg_a: a parent configuration
    :param cfg_b: a parent configuration
    :param cfg_c: a parent configuration
    :param a: weight for cfg_a
    :param b: weight for cfg_b
    :param c: weight for cfg_c
    """
    # attempt to normalize order, we prefer a==1.0
    if a != 1.0 and b == 1.0:  # swap a and b
      a, cfg_a, b, cfg_b = b, cfg_b, a, cfg_a
    if a != 1.0 and c == 1.0:  # swap a and c
      a, cfg_a, c, cfg_c = c, cfg_c, a, cfg_a

    # attempt to normalize order, we prefer b==-c
    if b < c:  # swap b and c
      b, cfg_b, c, cfg_c = c, cfg_c, b, cfg_b
    if b != -c and a == -c:  # swap a and c
      a, cfg_a, c, cfg_c = c, cfg_c, a, cfg_a

    if a == 1.0 and b == -c:
      self.copy_value(cfg_a, cfg)
      self.add_difference(cfg, b, cfg_b, cfg_c)  # TODO inline this logic?
    else:
      # TODO: should handle more cases
      self.op1_randomize(cfg)

  def add_difference(self, cfg_dst, scale, cfg_b, cfg_c):
    """
    add the difference cfg_b-cfg_c to cfg_dst

    this is the key operation used in differential evolution
    and some simplex techniques

    this operation is not possible in general with complex parameters but
    we make an attempt to "fake" it
    """
    if not self.same_value(cfg_b, cfg_c):
      self.op1_randomize(cfg_dst)

  @abc.abstractmethod
  def op1_randomize(self, config):
    """
    randomize this value without taking into account the current position
    :param config: the configuration to be changed
    """
    pass

  @abc.abstractmethod
  def seed_value(self):
    """some legal value of this parameter (for creating initial configs)"""
    return


class BooleanParameter(ComplexParameter):
  def manipulators(self, config):
    return [self.op1_flip]

  def get_value(self, config):
    return self._get(config)

  def set_value(self, config, value):
    self._set(config, value)

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration randomly

    :param config: the configuration to be changed
    """
    self._set(config, self.seed_value())

  def seed_value(self):
    return random.choice((True, False))

  def op1_flip(self, config):
    """
    Flip this parameter's value in a configuration

    :param config: the configuration to be changed
    """
    self._set(config, not self._get(config))

  def search_space_size(self):
    return 2

  def op3_swarm(self, cfg, cfg1, cfg2, c=1, c1=0.5,
                c2=0.5, velocity=0, *args, **kwargs):
    """
    Simulates a single update step in particle swarm optimization by updating
    the current position and returning a new velocity.

    The new velocity is given by

    .. math:: c*velocity + r1*c1*(cfg1-cfg) + r2*c2*(cfg2-cfg)

    where r1 and r2 are random values between 0 and 1

    The new current position is randomly chosen based on the new velocity

    :param cfg: the configuration to be changed. Represents the current position
    :param cfg1: a configuration to shift towards. Should be the local best position
    :param cfg2: a configuration to shift towards. Should be the global best position
    :param c: the weight of the current velocity
    :param c1: weight of cfg1
    :param c2: weight of cfg2
    :param velocity: the old velocity
    :param args:
    :param kwargs:
    :return: the new velocity, a float

    """
    v = velocity * c + (self.get_value(cfg1) - self.get_value(
        cfg)) * c1 * random.random() + (self.get_value(
        cfg2) - self.get_value(cfg)) * c2 * random.random()
    # Map velocity to continuous space with sigmoid
    s = old_div(1, (1 + numpy.exp(-v)))
    # Decide position randomly
    p = (s - random.random()) > 0
    self.set_value(cfg, p)
    return v


class SwitchParameter(ComplexParameter):
  """
  A parameter representing an unordered collection of options with no implied
  correlation between the choices. The choices are range(option_count)
  """

  def __init__(self, name, option_count):
    self.option_count = option_count
    super(SwitchParameter, self).__init__(name)

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration to a random value

    :param config: the configuration to be changed
    """
    self._set(config, random.randrange(self.option_count))

  def seed_value(self):
    return random.randrange(self.option_count)

  def search_space_size(self):
    return max(1, self.option_count)


class EnumParameter(ComplexParameter):
  """
  same as a SwitchParameter but choices are taken from an arbitrarily typed list
  """

  def __init__(self, name, options):
    super(EnumParameter, self).__init__(name)
    self.options = list(options)

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration to a random value

    :param config: the configuration to be changed
    """
    self._set(config, random.choice(self.options))

  def seed_value(self):
    return random.choice(self.options)

  def search_space_size(self):
    return max(1, len(self.options))


class PermutationParameter(ComplexParameter):
  """
  A parameter representing a permutation (or ordering) as a list of items
  """
  def __init__(self, name, items):
    super(PermutationParameter, self).__init__(name)
    self._items = list(items)
    self.size = len(items)

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration to a random value

    :param config: the configuration to be changed
    """
    random.shuffle(self._get(config))
    self.normalize(config)

  def op1_small_random_change(self, config, p=0.25):
    """
    Iterates through the list and probabilistically swaps each element with the
    next element

    :param p: probability of swapping an element with the next element
    :param config: the configuration to be changed
    """
    cfg_item = self._get(config)
    for i in range(1, len(cfg_item)):
      if random.random() < p:
        # swap
        cfg_item[i - 1], cfg_item[i] = cfg_item[i], cfg_item[i - 1]
    self.normalize(config)

  def seed_value(self):
    return list(self._items)  # copy

  def manipulators(self, config):
    return [self.op1_randomize, self.op1_small_random_change]

  def get_value(self, config):
    return self._get(config)

  def set_value(self, config, value):
    self._set(config, value)

  def search_space_size(self):
    return math.factorial(max(1, len(self._items)))

  def op3_cross(self, cfg, cfg1, cfg2, xchoice='op3_cross_OX1', strength=0.3,
                *args, **kwargs):
    """
    Calls the crossover operator specified by xchoice
    Passes argument d = strength*(size of the permutation)

    :param cfg: the configuration to be changed
    :param cfg1: a parent configuration
    :param cfg2: a parent configuration
    :param xchoice: string specifying which crossover operator to use (should start with op3_cross prefix)
    :param strength: the strength of the crossover
    """
    dd = int(round(self.size * strength))
    if dd < 1:
      log.warning('Crossover length too small. Cannot create new solution.')
    if dd >= self.size:
      log.warning('Crossover length too big. Cannot create new solution.')
    getattr(self, xchoice)(cfg, cfg1, cfg2, d=dd, *args, **kwargs)

  def op3_swarm(self, cfg, cfg1, cfg2, xchoice='op3_cross_OX1', c=0.5,
                c1=0.5, c2=0.5, strength=0.3, velocity=0, *args, **kwargs):
    """
    Replacement for particle swarm optimization iterative step for permutations.
    Given a target cfg and 2 parent cfgs, probabilistically performs an
    op3_cross with one of the 2 parents.

    :param cfg: the configuration to be changed. Represents the current position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param xchoice: which crossover operator should be used
    :param c: the probability of not performing a crossover
    :param c1: the probability of performing a crossover with cfg1 (if a
     crossover is performed)
    :param c2: unused
    :param strength: the strength of the crossover
    :param velocity: the old velocity - unused
    """
    if random.uniform(0, 1) > c:
      if random.uniform(0, 1) < c1:
        # Select crossover operator
        self.op3_cross(cfg, cfg, cfg1, xchoice, strength)
      else:
        self.op3_cross(cfg, cfg, cfg2, xchoice, strength)

  # swap-based operators
  def op2_random_swap(self, cfg, cfg1, *args, **kwargs):
    """
    Swap a random pair of items in cfg1 and save the result into cfg

    :param cfg: the configuration to be changed
    :param cfg1: the configuration whose PermutationParameter's elements are
     swapped and copied into cfg
    """
    p = self.get_value(cfg1)[:]
    r = random.randint(0, len(p) - 1)
    s = random.randint(0, len(p) - 1)
    v1 = p[r]
    v2 = p[s]
    p[r] = v2
    p[s] = v1
    self.set_value(cfg, p)

  def op2_random_invert(self, cfg, cfg1, strength=0.3, *args, **kwargs):
    """
    Reverse the ordering of a random subsection of size d in cfg1 and save the
    result in cfg where d = strength*total-size

    :param cfg: the configuration to be changed
    :param cfg1: the configuration whose PermutationParameter is inverted
    :param strength: the size of the reversed subsection as a fraction of the
     total size
    """
    p = self.get_value(cfg1)[:]
    d = int(round(len(p) * strength))
    r = random.randint(0, len(p) - d)
    subpath = p[r:r + d][:]
    subpath.reverse()
    p[r:r + d] = subpath
    self.set_value(cfg, p)

  # Crossover operators
  def op3_cross_PX(self, cfg, cfg1, cfg2, d=0):
    """
    Partition crossover (Whitley 2009?)

    Chooses a random cut point and reorders elements in cfg1 up to the cut point
    according to their order in cfg2.

    Saves the result in cfg

    :param cfg: the configuration to be changed
    :param cfg1: the first parent configuration. The "base" configuration
    :param cfg2: the second parent configuration. Is "crossed into" cfg1
    :param d: unused
    """
    p1 = self.get_value(cfg1)
    p2 = self.get_value(cfg2)
    c1 = random.randint(2, len(p1))
    self.set_value(cfg, sorted(p1[:c1], key=lambda x: p2.index(x)) + p1[c1:])

  def op3_cross_PMX(self, cfg, cfg1, cfg2, d=0):
    """
    Partially-mapped crossover Goldberg & Lingle (1985)

    Replaces a random section of cfg1 with the corresponding section in cfg2.
    Displaced elements in cfg1 are moved to the old position of the elements
    displacing them

    :param cfg: the configuration to be changed
    :param cfg1: the first parent configuration. The "base" configuration
    :param cfg2: the second parent configuration. Is "crossed into" cfg1
    :param d: the size of the crossover
    """
    if d == 0:
      d = max(1, int(round(self.size * 0.3))) # default to 1/3 of permutation size
    p1 = self.get_value(cfg1)[:]
    p2 = self.get_value(cfg2)[:]

    r = random.randint(0, len(p1) - d)

    c1 = p1[r:r + d]
    c2 = p2[r:r + d]

    # get new permutation by crossing over a section of p2 onto p1
    pnew = self.get_value(cfg1)[:]
    pnew[r:r + d] = c2
    # fix conflicts by taking displaced elements in crossed over section
    # displaced = (elements x in c1 where x does not have corresponding value in c2)
    # and putting them where the value that displaced them was

    #candidates for displacement
    candidate_indices = set(list(range(r)) + list(range(r+d, len(p1))))
    # Check through displaced elements to find values to swap conflicts to
    while c1 != []:
      n = c1[0]
      #try to match up a value in c1 to the equivalent value in c2
      while c2[0] in c1:
        if n == c2[0]:
          # already match up
          break
        # find position idx of c2[0] in c1
        link_idx = c1.index(c2[0])
        # get value of c2 at idx
        link = c2[link_idx]
        # remove c2[idx] and c1[idx] since they match up when we swap c2[0] with c2[idx] (this avoids an infinite loop)
        del c2[link_idx]
        del c1[link_idx]
        # swap new value into c2[0]
        c2[0] = link

      if n != c2[0]:
        # first check if we can swap in the crossed over section still
        if n in c2:
          c2[c2.index(n)] = c2[0]
        else:
          # assign first instance of c2[0] outside of the crossed over section in pnew to c1[0]
          for idx in candidate_indices:
            if pnew[idx] == c2[0]:
              pnew[idx] = c1[0]
              candidate_indices.remove(idx) # make sure we don't override this value now
              break
      # remove first elements
      del c1[0]
      del c2[0]
    self.set_value(cfg, pnew)

  def op3_cross_CX(self, cfg, cfg1, cfg2, d=0):
    """
    Implementation of a cyclic crossover.

    Repeatedly replaces elements of cfg1 with the element at the same index in
    cfg2. This is done until a cycle is reached and cfg1 is valid again. The
    initial replacement is random.

    Saves the result in cfg.

    :param cfg: the configuration to be changed
    :param cfg1: the first parent configuration. The "base" configuration
    :param cfg2: the second parent configuration. Is "crossed into" cfg1
    :param d: unused
    """
    p1 = self.get_value(cfg1)
    p2 = self.get_value(cfg2)
    p = p1[:]

    s = random.randint(0, len(p1) - 1)
    i = s
    indices = set()

    while len(indices) < len(p1): # should never exceed this
      indices.add(i)
      val = p1[i]
      i = p2.index(val)
      # deal with duplicate values
      while i in indices:
        if i == s:
          break
        i = p2[i+1:].index(val) + i + 1
      if i == s:
        break

    for j in indices:
      p[j] = p2[j]

    self.set_value(cfg, p)

  def op3_cross_OX1(self, cfg, cfg1, cfg2, d=0):
    """
    Ordered Crossover (Davis 1985)

    Exchanges a subpath from cfg2 into cfg1 while maintaining the order of the
    remaining elements in cfg1.

    Saves the result in cfg.

    :param cfg: the configuration to be changed
    :param cfg1: the first parent configuration. The "base" configuration
    :param cfg2: the second parent configuration. Is "crossed into" cfg1
    :param d: size of the exchanged subpath
    """
    if d == 0:
      d = max(1, int(round(self.size * 0.3))) # default to 1/3 of permutation size
    p1 = self.get_value(cfg1)
    p2 = self.get_value(cfg2)
    c1 = p1[:]
    c2 = p2[:]
    # Randomly find cut points
    r = random.randint(0, len(
        p1) - d)  # Todo: treat path as circle i.e. allow cross-boundary cuts
    [c1.remove(i) for i in p2[r:int(r + d)]]
    self.set_value(cfg, c1[:r] + p2[r:r + d] + c1[r:])

  def op3_cross_OX3(self, cfg, cfg1, cfg2, d=0):
    """
    Ordered crossover variation 3 (Deep 2010)

    Same as op3_cross_OX1, except the parents have different cut points for
    their subpaths

    :param cfg: the configuration to be changed
    :param cfg1: the first parent configuration. The "base" configuration
    :param cfg2: the second parent configuration. Is "crossed into" cfg1
    :param d: size of the exchanged subpath
    """
    if d == 0:
      d = max(1, int(round(self.size * 0.3))) # default to 1/3 of permutation size
    p1 = self.get_value(cfg1)
    p2 = self.get_value(cfg2)
    c1 = p1[:]
    c2 = p2[:]
    # Randomly find cut points
    # Todo: treat path as circle i.e. allow cross-boundary cuts
    r1 = random.randint(0, len(p1) - d)
    r2 = random.randint(0, len(p1) - d)
    [c1.remove(i) for i in p2[r2:r2 + d]]
    self.set_value(cfg, c1[:r1] + p2[r2:r2 + d] + c1[r1:])

  def search_space_size(self):
    return math.factorial(max(1, len(self._items)))


class ScheduleParameter(PermutationParameter):
  def __init__(self, name, items, deps):
    super(ScheduleParameter, self).__init__(name, items)
    self.deps = dict((k, set(v)) for k, v in list(deps.items()))
    log.debug("ScheduleParameter(%s, %s, %s)", repr(name), repr(items),
              repr(deps))
    self._expand_deps()

  def _expand_deps(self):
    """expand self.deps to include recursive dependencies"""
    fixed_point = False
    while not fixed_point:
      fixed_point = True
      for k in list(self.deps.keys()):
        oldlen = len(self.deps[k])
        for dep in list(self.deps[k]):
          if dep in self.deps:
            self.deps[k].update(self.deps[dep])
        if oldlen != len(self.deps[k]):
          fixed_point = False

    # verify schedule is valid
    items = set(self._items)
    for k, v in list(self.deps.items()):
      if k in v:
        raise Exception("ScheduleParameter('%s') cycle: %s depends on itself" %
                        (self.name, k))

      if v - items:
        raise Exception("ScheduleParameter('%s'): %s is unknown" %
                        (self.name, v - items))

    if set(self.deps.keys()) - items:
      raise Exception("ScheduleParameter('%s'): %s is unknown" %
                      (self.name, set(self.deps.keys()) - items))

  def is_topologically_sorted(self, values):
    used = set()
    for v in values:
      if v in self.deps and self.deps[v].union(used):
        return False
      used.add(v)
    return True

  def topologically_sorted_depth_first(self, values):
    """faster but not stable enough"""
    if self.is_topologically_sorted(values):
      return values
    sorted_values = []
    used = set()
    deps = dict((k, sorted(v, key=values.index, reverse=True))
                for k, v in list(self.deps.items()))

    def visit(v):
      if v in used:
        return
      if v in deps:
        for dv in deps[v]:
          visit(dv)
      used.add(v)
      sorted_values.append(v)

    for v in reversed(values):
      visit(v)
    return list(reversed(sorted_values))

  def topologically_sorted(self, values):
    if self.is_topologically_sorted(values):
      return values
    deps = copy.deepcopy(self.deps)
    queue = collections.deque(reversed(values))
    sorted_values = []
    while queue:
      v = queue.popleft()
      if v in deps and deps[v]:
        queue.append(v)
      else:
        for k, d in list(deps.items()):
          d.discard(v)
          if not d:
            del deps[k]
        sorted_values.append(v)

    return list(reversed(sorted_values))

  def normalize(self, cfg):
    self._set(cfg, self.topologically_sorted(self._get(cfg)))


class SelectorParameter(ComplexParameter):
  def __init__(self, name, choices, max_cutoff,
               order_class=PermutationParameter,
               offset_class=LogIntegerParameter):
    super(SelectorParameter, self).__init__(name)
    self.choices = choices
    self.max_cutoff = max_cutoff
    self.order_param = order_class('{0}/order'.format(name), choices)
    self.offset_params = [
        offset_class('{0}/offsets/{1}'.format(name, i), 0, max_cutoff)
        for i in range(len(choices) - 1)]

  def sub_parameters(self):
    return [self.order_param] + self.offset_params

  def seed_value(self):
    return {'order': self.order_param.seed_value(),
            'offsets': [co.seed_value() for co in self.offset_params]}

  def op1_randomize(self, config):
    random.choice(self.sub_parameters()).op1_randomize(config)

  def selector_iter(self, config):
    """
    yield (cutoff, choice) pairs
    cutoff will be None on the first value
    """
    order = config[self.name]['order']
    yield (None, order[0])
    cutoff = 0
    for n, offset in enumerate(config[self.name]['offsets']):
      if offset > 0:
        cutoff += offset
        yield cutoff, order[n + 1]


class ParameterArray(ComplexParameter):
  """
  Represents an array of Parameters
  """
  def __init__(self, name, count, element_type, *args, **kwargs):
    super(ParameterArray, self).__init__(name)
    self.count = count

    self.sub_params = [
        element_type('{0}/{1}'.format(name, i), *args[i], **kwargs[i])
        for i in range(count)]

  def sub_parameters(self):
    return self.sub_params

  def seed_value(self):
    return [p.seed_value() for p in self.sub_params]

  def op1_randomize(self, config):
    """
    randomly selects a sub-parameter and randomizes it

    :param config: the configuration to be changed
    """
    random.choice(self.sub_parameters()).op1_randomize(config)


class BooleanParameterArray(ParameterArray):
  """
  Represents an array of BooleanParameters - currently unimplimented
  """
  def __init__(self, name, count):
    super(BooleanParameterArray, self).__init__(name, count, BooleanParameter)

  def op3_swarm(self, cfg, cfg1, cfg2, *args, **kwargs):
    # TODO
    pass

  def op3_cross(self, cfg, cfg1, cfg2, *args, **kwargs):
    # TODO
    pass


class IntegerParameterArray(ParameterArray):
  """
  Represents an array of IntegerParameters - currently unimplemented
  """
  def __init__(self, name, min_values, max_values):
    assert len(min_values) == len(max_values)
    super(IntegerParameterArray, self).__init__(name, len(min_values),
                                                IntegerParameter,
                                                min_value=min_values,
                                                max_value=max_values)

  def op3_swarm(self, cfg, cfg1, cfg2, *args, **kwargs):
    # TODO
    pass

  def op3_cross(self, cfg, cfg1, cfg2, *args, **kwargs):
    # TODO
    pass


class Array(ComplexParameter):
  """
  An interface for parameters representing an array of values.
  """
  # TODO: constraints? (upper & lower bound etc)
  def __init__(self, name, size):
    super(Array, self).__init__(name)
    self.size = size

  def op3_cross(self, cfg, cfg1, cfg2, strength=0.3, *args, **kwargs):
    """
    Crosses two arrays by replacing a random subsection of cfg1 with the
    corresponding subsection of cfg2.The size of the chunk is a fixed fraction
    of the total length, given by the strength

    Behaves like a specialized 2-point crossover, where the first cut point is
    random and the second cut is a set distance after.

    :param cfg: the configuration to be changed
    :param cfg1: the configuration being inserted into
    :param cfg2: the configuration being inserted
    :param strength: the size of the crossover, as a fraction of total array
     length
    """
    d = int(round(self.size * strength))
    if d < 1:
      log.debug('Crossover length too small. Cannot create new solution.')
    if d >= self.size:
      log.debug('Crossover length too big. Cannot create new solution.')
    p1 = self.get_value(cfg1)
    p2 = self.get_value(cfg2)
    r = random.randint(0, len(
        p1) - d)  # Todo: treat path as circle i.e. allow cross-boundary cuts
    p = numpy.concatenate([p1[:r], p2[r:r + d], p1[r + d:]])
    self.set_value(cfg, p)

  def op3_swarm(self, cfg, cfg1, cfg2, c=1, c1=0.5,
                c2=0.5, velocity=0, strength=0.3, *args, **kwargs):
    """
    Replacement for a particle swarm optimization iterative step for arrays.
    Given a target cfg and 2 parent cfgs, probabilistically performs an
    :py:meth:`op3_cross` with one of the 2 parents.

    :param cfg: the configuration to be changed. Represents the cfg position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param c: the probability of not performing a crossover
    :param c1: the probability of performing a crossover with cfg1 (if a
     crossover is performed)
    :param c2: unused
    :param velocity: the old velocity - unused
    :param strength: the strength of the crossover
    """
    if random.uniform(0, 1) > c:
      if random.uniform(0, 1) < c1:
        # Select crossover operator
        self.op3_cross(cfg, cfg, cfg1, strength)
      else:
        self.op3_cross(cfg, cfg, cfg2, strength)

  def get_value(self, config):
    return self._get(config)

  def set_value(self, config, value):
    self._set(config, value)


class BooleanArray(Array):
  """
  Represents an array of boolean values which are either 0 or 1
  """
  def op3_swarm_parallel(self, cfg, cfg1, cfg2, c=1,
                         c1=0.5, c2=0.5, velocities=0):
    """
    Simulates a single particle swarm optimization step for each element in the
    array by updating each position and returning an array of new velocities.

    The new velocities are given by

    .. math:: c*velocity + r1*c1*(cfg1-cfg) + r2*c2*(cfg2-cfg)

    where r1 and r2 are random values between 0 and 1. In each iteration, r1 and
    r2 are constant across array elements

    The new cfg positions are randomly chosen based on the new velocities

    :param cfg: the configuration to be changed. This represents the current
     position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param c: the weight of the current velocities
    :param c1: weight of cfg1
    :param c2: weight of cfg2
    :param velocities: the current velocities
    :return: a numpy array of new velocities
    """
    vs = velocities * c + (self.get_value(cfg1) - self.get_value(
        cfg)) * c1 * random.random() + (self.get_value(
            cfg2) - self.get_value(cfg)) * c2 * random.random()
    # Map velocity to continuous space with sigmoid
    ss = old_div(1, (1 + numpy.exp(-vs)))
    # Decide position randomly
    ps = (ss - numpy.random.rand(1, self.size)) > 0
    self.set_value(cfg, ps)
    return vs

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration randomly

    :param config: the configuration to be changed
    """
    value = numpy.random.rand(1, self.size) > 0.5
    self._set(config, value)

  def seed_value(self):
    return numpy.random.rand(1, self.size) > 0.5


class FloatArray(Array):
  """
  Represents an array of float values
  """
  def __init__(self, name, size, fmax, fmin):
    super(FloatArray, self).__init__(name, size)
    self.fmax = fmax
    self.fmin = fmin

  def op1_randomize(self, config):
    """
    Set this parameter's value in a configuration randomly

    :param config: the configuration to be changed
    """
    value = numpy.random.rand(1, self.size) * (
        self.fmax - self.fmin) + self.fmin
    self._set(config, value)

  def seed_value(self):
    value = numpy.random.rand(1, self.size) * (
        self.fmax - self.fmin) + self.fmin
    return value

  def op3_swarm_parallel(self, cfg, cfg1, cfg2, c=1,
                         c1=0.5, c2=0.5, velocities=0):
    """
    Simulates a single particle swarm optimization step for each element in the
    array by updating the each position and returning an array of new velocities

    The new velocity is given by

    .. math:: c*velocity + r1*c1*(cfg1-cfg) + r2*c2*(cfg2-cfg)

    where r1 and r2 are random values between 0 and 1. In each iteration, r1 and
    r2 are constant across array elements

    The new cfg positions are randomly chosen based on the new velocities

    :param cfg: the configuration to be changed. This represents the current
     position
    :param cfg1: a configuration to shift towards. Should be the local best
     position
    :param cfg2: a configuration to shift towards. Should be the global best
     position
    :param c: the weight of the cfg velocities
    :param c1: weight of cfg1
    :param c2: weight of cfg2
    :param velocities: the cfg velocities
    :return: a numpy array of new velocities
    """
    vs = velocities * c + (self.get_value(cfg1) - self.get_value(
        cfg)) * c1 * random.random() + (self.get_value(
        cfg2) - self.get_value(cfg)) * c2 * random.random()
    p = self.get_value(cfg) + vs
    p[p > self.fmax] = self.fmax
    p[p < self.fmin] = self.fmin
    self.set_value(cfg, p)
    return vs


##################

class ManipulatorProxy(object):
  """
  wrapper around configuration manipulator and config pair
  """

  def __init__(self, manipulator, cfg):
    self.cfg = cfg
    self.manipulator = manipulator
    self.params = manipulator.parameters_dict(self.cfg)

  def keys(self):
    return list(self.params.keys())

  def __getitem__(self, k):
    return ParameterProxy(self.params[k], self.cfg)


class ParameterProxy(object):
  """
  wrapper aint parameter and config pair, adds config
  as first argument to all method calls to parameter
  """

  def __init__(self, param, cfg):
    self.cfg = cfg
    self.param = param

  def __getattr__(self, key):
    """equivalent of self.param.key(self.cfg, ...)"""
    member = getattr(self.param, key)

    def param_method_proxy(*args, **kwargs):
      return member(self.cfg, *args, **kwargs)

    if callable(member):
      return param_method_proxy
    else:
      # we should only hit this for key == 'name'
      return member


# Inspection Methods
def operators(param, num_parents):
  """
  Return a list of operators for the given parameter that take the specified
  number of input configurations

  :param param: a Parameter class
  :param num_parents: a String specifying number of inputs required by the operator.
    should be one of '1', '2', '3', '4', or 'n'
  """
  ops = []
  methods = inspect.getmembers(param, inspect.ismethod)
  for m in methods:
    name, obj = m
    if is_operator(name, num_parents):
      ops.append(name)
  return ops

def composable_operators(param, min_num_parents):
  """
  Return a list of operators for the given parameter that can be programatically composed
  with a composable technique generating min_num_parents.

  Programatically composable operators have no non-cfg arguments

  :param param: a Parameter class
  :param min_num_parents: the minimum number of parents passed to the operator
  """
  if min_num_parents < 1:
    return []

  allowed_num_parents = ['n']
  for i in range(1,5):
    if i > min_num_parents:
      break
    allowed_num_parents.append(str(i))

  ops = []
  methods = inspect.getmembers(param, inspect.ismethod)
  for m in methods:
    name, obj = m
    argspec = inspect.getargspec(obj)
    numargs = len(argspec.args) - (len(argspec.defaults) if argspec.defaults else 0)
    for num_parents in allowed_num_parents:
      if is_operator(name, num_parents):
        if num_parents == 'n':
          if numargs == 3: # self, cfg, cfgs
            ops.append(name)
        else:
          if numargs == (1 + int(num_parents)):
            ops.append(name)
        break
  return ops


def is_operator(name, num_parents):
  """
  Tells whether a method is an operator taking in the specified number of inputs
  from the method name

  :param name: the method name
  :param num_parents: a String specifying number of inputs required by the operator.
    should be one of '1', '2', '3', '4', or 'n'
  """
  return ('op' + num_parents + '_') == name[:4]

def all_operators():
  """
  Return a dictionary mapping from parameter names to lists of operator function
  names
  """
  ops = {}
  for p in all_params():
    name, obj = p
    all_ops = []
    for num in ['1', '2', '3', '4', 'n']:
      all_ops += operators(obj, num)
    ops[name] = all_ops
  return ops

def all_params():
  params = inspect.getmembers(sys.modules[__name__], lambda x: inspect.isclass(
    x) and x.__module__ == __name__ and issubclass(x, Parameter))
  return params




from __future__ import division
from builtins import str
from builtins import map
from builtins import filter
from builtins import range
from builtins import object
from past.utils import old_div
from functools import cmp_to_key
import abc
import logging
import math
from collections import defaultdict
from fn import _
from fn.iters import map, filter
from .manipulator import Parameter
from .metatechniques import RecyclingMetaTechnique
from .technique import SequentialSearchTechnique, register

log = logging.getLogger(__name__)


class SimplexTechnique(SequentialSearchTechnique):
  """
  Base class with utility functions common
  to simplex type methods
  """

  def __init__(self, seed_cfg=None, *args, **kwargs):
    super(SimplexTechnique, self).__init__(*args, **kwargs)
    self.centroid = None
    self.last_simplex_points = None
    self.seed_cfg = seed_cfg
    self.simplex_points = []

  def calculate_centroid(self):
    """
    average of all the PrimitiveParameters in self.simplex_points
    ComplexParameters are copied from self.simplex_points[0]
    """
    sums = defaultdict(float)
    counts = defaultdict(int)

    for config in self.simplex_points:
      cfg = config.data
      for param in self.manipulator.parameters(cfg):
        if param.is_primitive():
          sums[param.name] += param.get_unit_value(cfg)
          counts[param.name] += 1

    centroid = self.manipulator.copy(self.simplex_points[0].data)
    for param in self.manipulator.parameters(centroid):
      if param.is_primitive():
        param.set_unit_value(centroid,
                             old_div(sums[param.name], float(counts[param.name])))

    return centroid

  def cfg_to_str(self, cfg):
    params = list(filter(Parameter.is_primitive,
                         self.manipulator.parameters(cfg)))
    params.sort(key=_.name)
    return str(tuple([x.get_unit_value(cfg) for x in params]))

  def debug_log(self):
    for i, config in enumerate(self.simplex_points):
      log.debug("simplex_points[%d] = %s", i, self.cfg_to_str(config.data))
    if self.centroid:
      log.debug("centroid = %s", self.cfg_to_str(self.centroid))

  def linear_point(self, p1, p2, scale):
    """
    return a point on the line passing between p1 and p2 at position scale
    such that p1 + scale*(p1 - p2)
    """
    return self.manipulator.linear_config(1.0, p1, scale, p1, -scale, p2)

  def convergence_criterea(self):
    """True will cause the simplex method to stop"""
    if self.rounds_since_novel_request > 3 * len(self.simplex_points) + 1:
      return True
    if self.last_simplex_points == self.simplex_points:
      return True
    self.last_simplex_points = list(self.simplex_points)
    return False

  def initial_simplex_seed(self):
    """
    return a point to base the initial simplex on
    """
    if self.seed_cfg is not None:
      return self.seed_cfg
    return self.manipulator.random()

  @abc.abstractmethod
  def initial_simplex(self):
    """
    return a initial list of configurations
    """
    return []


class RandomInitialMixin(object):
  """
  start with random initial simplex
  """

  def initial_simplex(self):
    # we implicitly assume number of parameters is fixed here, however
    # it will work if it isn't (simplex size is undefined)
    cfg0 = self.initial_simplex_seed()
    params = self.manipulator.parameters(cfg0)
    return [cfg0] + [self.manipulator.random()
                     for p in params
                     if p.is_primitive()]


class RightInitialMixin(object):
  """
  start with random initial right triangle like simplex
  """

  def __init__(self, initial_unit_edge_length=0.1, *args, **kwargs):
    assert initial_unit_edge_length <= 0.5
    self.initial_unit_edge_length = initial_unit_edge_length
    super(RightInitialMixin, self).__init__(*args, **kwargs)

  def initial_simplex(self):
    cfg0 = self.initial_simplex_seed()
    simplex = [cfg0]
    params = self.manipulator.parameters(cfg0)
    params = [x for x in params if x.is_primitive()]
    for p in params:
      simplex.append(self.manipulator.copy(cfg0))
      v = p.get_unit_value(simplex[-1])
      if v <= 0.5:
        v += self.initial_unit_edge_length
      else:
        v -= self.initial_unit_edge_length
      p.set_unit_value(simplex[-1], v)
    return simplex


class RegularInitialMixin(object):
  """
  start with random initial regular simplex (all edges equal length)
  """

  def __init__(self, initial_unit_edge_length=0.1, *args, **kwargs):
    assert initial_unit_edge_length <= 0.5
    self.initial_unit_edge_length = initial_unit_edge_length
    super(RegularInitialMixin, self).__init__(*args, **kwargs)

  def initial_simplex(self):
    cfg0 = self.initial_simplex_seed()
    simplex = [cfg0]
    params = self.manipulator.parameters(cfg0)
    params = list([x for x in params if x.is_primitive()])
    if len(params) == 0:
      return simplex

    q = ((old_div((math.sqrt(len(params) + 1.0) - 1.0), (len(params) * math.sqrt(2.0))))
         * self.initial_unit_edge_length)
    p = q + ((old_div(1.0, math.sqrt(2.0))) * self.initial_unit_edge_length)

    base = [x.get_unit_value(cfg0) for x in params]
    for j in range(len(base)):
      if max(p, q) + base[j] > 1.0:
        #flip this dimension as we would overflow our [0,1] bounds
        base[j] *= -1.0

    for i in range(len(params)):
      simplex.append(self.manipulator.copy(cfg0))
      params[i].set_unit_value(simplex[-1], abs(base[i] + p))
      for j in range(i + 1, len(params)):
        params[j].set_unit_value(simplex[-1], abs(base[i] + q))

    return simplex


class NelderMead(SimplexTechnique):
  """
  Nelder-Mead downhill simplex method.

  Based on description of method on page 82 of
  'Noisy Optimization With Evolution Strategies' by Dirk V. Arnold.

  We set alpha=2.0 by default instead of the often recommended alpha=1.0 to
  avoid a common degenerate case, where the volume of the simplex becomes zero.
  This is easiest to see with a single parameter. Let the simplex points
  be x0,x1.  Let the centroid be c=(x0+x1)/2.0 and the reflection point be:
  reflection = c + alpha*(c-x1) = (x0+x1)*(1+alpha)/2 - x1
  The problem is, if we set alpha = 1.0, then the x1's cancel out and the
  reflection point becomes just reflection=x0, which also happens to be the
  second best point, meaning we will use it.  So in a single step of the
  algorithm the simplex becomes singular.
  """

  def __init__(self,
               alpha=2.0,
               gamma=2.0,
               beta=0.5,
               sigma=0.5,
               *args, **kwargs):
    self.alpha = alpha
    self.gamma = gamma
    self.beta = beta
    self.sigma = sigma
    super(NelderMead, self).__init__(*args, **kwargs)

  @classmethod
  def get_hyper_parameters(cls):
    return ['alpha', 'gamma', 'beta', 'sigma']


  def main_generator(self):
    objective = self.objective
    driver = self.driver

    # test the entire initial simplex
    self.simplex_points = list(map(driver.get_configuration,
                                   self.initial_simplex()))

    if len(self.simplex_points) <= 1:
      log.warning("only 1 point in simplex, will not use %s", self.name)
      return

    log.debug("initial points")
    for p in self.simplex_points:
      self.yield_nonblocking(p)
    yield None  # wait until results are ready

    while not self.convergence_criterea():
      # next steps assume this ordering
      self.simplex_points.sort(key=cmp_to_key(objective.compare))
      # set limit from worst point
      self.limit = objective.limit_from_config(self.simplex_points[-1])
      self.centroid = self.calculate_centroid()
      if log.isEnabledFor(logging.DEBUG):
        self.debug_log()

      reflection = self.reflection_point()
      yield reflection

      if objective.lt(reflection, self.simplex_points[0]):
        #expansion case
        expansion = self.expansion_point(reflection)
        yield expansion

        if objective.lt(expansion, reflection):
          log.debug("using expansion point")
          self.simplex_points[-1] = expansion
        else:
          log.debug("using reflection point (considered expansion)")
          self.simplex_points[-1] = reflection

      elif objective.lt(reflection, self.simplex_points[1]):
        #reflection case
        log.debug("using reflection point")
        self.simplex_points[-1] = reflection
      else:
        # contraction case
        if objective.lte(reflection, self.simplex_points[-1]):
          # outside contraction
          contract_base = reflection
        else:
          # inside contraction
          contract_base = self.simplex_points[-1]

        contraction = self.contraction_point(contract_base)
        yield contraction

        if objective.lte(contraction, contract_base):
          log.debug("using contraction point")
          self.simplex_points[-1] = contraction
        else:
          #reduction case
          log.debug("performing shrink reduction")
          self.perform_shrink_reduction()
          for p in self.simplex_points:
            self.yield_nonblocking(p)
          yield None  # wait until results are ready

  def reflection_point(self):
    """
    reflect worst point across centroid
    """
    return self.driver.get_configuration(
        self.linear_point(self.centroid,
                          self.simplex_points[-1].data,
                          self.alpha))

  def expansion_point(self, reflection):
    """
    reflect worst point across centroid more (by default 2x as much)
    """
    return self.driver.get_configuration(
        self.linear_point(self.centroid,
                          reflection.data,
                          -self.gamma))

  def contraction_point(self, contract_base):
    """
    reflect worst point across centroid less
    """
    return self.driver.get_configuration(
        self.linear_point(self.centroid,
                          contract_base.data,
                          -self.beta))

  def perform_shrink_reduction(self):
    """
    shrink the simplex in size by sigma=1/2 (default), moving it closer to the
    best point
    """
    for i in range(1, len(self.simplex_points)):
      self.simplex_points[i] = self.driver.get_configuration(
          self.linear_point(self.simplex_points[0].data,
                            self.simplex_points[i].data,
                            -self.sigma))


class Torczon(SimplexTechnique):
  """
  Torczon multi-directional search algorithm.

  Based on description of method on page 85 of
  'Noisy Optimization With Evolution Strategies' by Dirk V. Arnold.
  """

  def __init__(self,
               alpha=1.0,
               gamma=2.0,
               beta=0.5,
               *args, **kwargs):
    self.alpha = alpha
    self.gamma = gamma
    self.beta = beta
    super(Torczon, self).__init__(*args, **kwargs)

  @classmethod
  def get_hyper_parameters(cls):
    return ['alpha', 'gamma', 'beta']


  def main_generator(self):
    objective = self.objective
    driver = self.driver

    # test the entire initial simplex
    self.simplex_points = list(map(driver.get_configuration,
                                   self.initial_simplex()))
    if len(self.simplex_points) <= 1:
      log.warning("only 1 point in simplex, will not use %s", self.name)
      return

    log.debug("initial points")
    for p in self.simplex_points:
      self.yield_nonblocking(p)
    yield None  # wait until results are ready
    self.simplex_points.sort(cmp=objective.compare)

    while not self.convergence_criterea():
      # set limit from worst point
      self.limit = objective.limit_from_config(self.simplex_points[-1])

      if log.isEnabledFor(logging.DEBUG):
        self.debug_log()

      reflected = self.reflected_simplex()
      yield None  # wait until results are ready
      reflected.sort(cmp=objective.compare)

      # this next condition implies reflected[0] < simplex_points[0] since
      # reflected is sorted and contains simplex_points[0] (saves a db query)
      if reflected[0] is not self.simplex_points[0]:
        expanded = self.expanded_simplex()
        yield None  # wait until results are ready
        expanded.sort(cmp=objective.compare)

        if objective.lt(expanded[0], reflected[0]):
          log.debug("expansion performed")
          self.simplex_points = expanded
        else:
          log.debug("reflection performed")
          self.simplex_points = reflected
      else:
        contracted = self.contracted_simplex()
        yield None  # wait until results are ready
        contracted.sort(cmp=objective.compare)

        log.debug("contraction performed")
        self.simplex_points = contracted

  def scaled_simplex(self, scale):
    """
    assumes self.simplex_points[0] is best point and returns a new simplex
    reflected across self.simplex_points[0] by scale
    """
    simplex = list(self.simplex_points)  # shallow copy
    for i in range(1, len(simplex)):
      simplex[i] = self.driver.get_configuration(
          self.linear_point(simplex[0].data, simplex[i].data, scale))
      self.yield_nonblocking(simplex[i])
    return simplex

  def reflected_simplex(self):
    return self.scaled_simplex(self.alpha)

  def expanded_simplex(self):
    return self.scaled_simplex(self.gamma)

  def contracted_simplex(self):
    return self.scaled_simplex(-self.beta)


class RandomNelderMead(RandomInitialMixin, NelderMead):
  pass


class RightNelderMead(RightInitialMixin, NelderMead):
  pass


class RegularNelderMead(RegularInitialMixin, NelderMead):
  pass


class RandomTorczon(RandomInitialMixin, Torczon):
  pass


class RightTorczon(RightInitialMixin, Torczon):
  pass


class RegularTorczon(RegularInitialMixin, Torczon):
  pass


class MultiNelderMead(RecyclingMetaTechnique):
  def __init__(self):
    super(MultiNelderMead, self).__init__([RightNelderMead, RandomNelderMead,
                                           RegularNelderMead])


class MultiTorczon(RecyclingMetaTechnique):
  def __init__(self):
    super(MultiTorczon, self).__init__([RightTorczon, RandomTorczon,
                                        RegularTorczon])


register(RandomNelderMead())
register(RegularNelderMead())
register(RightNelderMead())
register(MultiNelderMead())
register(RandomTorczon())
register(RegularTorczon())
register(RightTorczon())
register(MultiTorczon())









#!/usr/bin/env python

from __future__ import absolute_import
if __name__ == '__main__':
  from . import adddeps

import argparse
import logging
import sys

import opentuner
from opentuner.resultsdb.models import *

log = logging.getLogger('opentuner.utils.compactdb')

argparser = argparse.ArgumentParser()
argparser.add_argument('database')
argparser.add_argument('--level', type=int, default=2)


def main(args):
  if '://' not in args.database:
    args.database = "sqlite:///" + args.database
  engine, Session = opentuner.resultsdb.connect(args.database)
  session = Session()

  config_count = session.query(Configuration).count()
  # result_count = session.query(Result).count()
  # desired_result_count = session.query(DesiredResult).count()

  if args.level >= 1:
    q = (session.query(Configuration)
         .filter(~Configuration.id.in_(session.query(Result.configuration_id)
                                       .filter_by(was_new_best=True)
                                       .subquery()))
         .filter(Configuration.data != None))

    log.info("%s: compacted %d of %d Configurations",
             args.database,
             q.update({'data': None}, False),
             config_count)
    session.commit()

  if args.level >= 2:
    session.execute('VACUUM;')
    session.commit()

  log.info('done')


if __name__ == '__main__':
  opentuner.tuningrunmain.init_logging()
  sys.exit(main(argparser.parse_args()))





from __future__ import absolute_import
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from .models import Base, _Meta
import logging
import time
from pprint import pprint

log = logging.getLogger(__name__)

DB_VERSION = "0.0"

if False:  # profiling of queries
  import atexit
  from sqlalchemy import event
  from collections import Counter
  from sqlalchemy.engine import Engine
  the_query_totals = Counter()

  @event.listens_for(Engine, "before_cursor_execute")
  def before_cursor_execute(conn, cursor, statement,
                            parameters, context, executemany):
      context._query_start_time = time.time()

  @event.listens_for(Engine, "after_cursor_execute")
  def after_cursor_execute(conn, cursor, statement,
                           parameters, context, executemany):
      total = time.time() - context._query_start_time
      the_query_totals[statement] += total

  @atexit.register
  def report():
    pprint(the_query_totals.most_common(10))


def connect(dbstr):
  engine = create_engine(dbstr, echo = False)
  connection = engine.connect()

  #handle case that the db was initialized before a version table existed yet
  if engine.dialect.has_table(connection, "program"):
    # if there are existing tables
    if not engine.dialect.has_table(connection, "_meta"):
      # if no version table, assume outdated db version and error
      connection.close()
      raise Exception("Your opentuner database is currently out of date. Save a back up and reinitialize")

  # else if we have the table already, make sure version matches
  if engine.dialect.has_table(connection, "_meta"):
    Session = scoped_session(sessionmaker(autocommit=False,
                                          autoflush=False,
                                          bind=engine))
    version = _Meta.get_version(Session)
    if not DB_VERSION == version:
      raise Exception('Your opentuner database version {} is out of date with the current version {}'.format(version, DB_VERSION))

  Base.metadata.create_all(engine)

  Session = scoped_session(sessionmaker(autocommit=False,
                                        autoflush=False,
                                        bind=engine))
  # mark database with current version
  _Meta.add_version(Session, DB_VERSION)
  Session.commit()

  return engine, Session




#!/usr/bin/python
try:
    from setuptools import setup
except ImportError:
    try:
        from setuptools.core import setup
    except ImportError:
        from distutils.core import setup

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rest')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

required = open('requirements.txt').read().splitlines()
required = [l.strip() for l in required
            if l.strip() and not l.strip().startswith('#')]

setup(
    name='opentuner',
    version='0.8.1',
    url='http://opentuner.org/',
    license='MIT',
    author='Jason Ansel',
    author_email='jansel@jansel.net',
    description='An extensible framework for program autotuning',
    long_description=read_md('README.md'),
    packages=['opentuner', 'opentuner.resultsdb', 'opentuner.utils',
              'opentuner.measurement', 'opentuner.search'],
    install_requires=required,
)



from __future__ import division
from past.builtins import cmp
from builtins import map
from past.utils import old_div
from builtins import object
import abc
import logging

from fn import _
from sqlalchemy import text

import opentuner
from opentuner.resultsdb.models import *
from future.utils import with_metaclass

log = logging.getLogger(__name__)


class SearchObjective(with_metaclass(abc.ABCMeta, object)):
  """
  delegates the comparison of results and configurations
  """

  @abc.abstractmethod
  def result_order_by_terms(self):
    """return database columns required to order by the objective"""
    return []

  @abc.abstractmethod
  def result_compare(self, result1, result2):
    """cmp() compatible comparison of resultsdb.models.Result"""
    return

  def config_compare(self, config1, config2):
    """cmp() compatible comparison of resultsdb.models.Configuration"""
    return self.result_compare(self.driver.results_query(config=config1).one(),
                               self.driver.results_query(config=config2).one())

  @abc.abstractmethod
  def result_relative(self, result1, result2):
    """return None, or a relative goodness of resultsdb.models.Result"""
    return

  def config_relative(self, config1, config2):
    """return None, or a relative goodness of resultsdb.models.Configuration"""
    return self.result_relative(self.driver.results_query(config=config1).one(),
                                self.driver.results_query(config=config2).one())


  def __init__(self):
    self.driver = None

  def set_driver(self, driver):
    self.driver = driver

  def result_order_by(self, q):
    return q.order_by(*self.result_order_by_terms())

  def compare(self, a, b):
    """cmp() compatible compare"""
    if isinstance(a, Configuration):
      return self.config_compare(a, b)
    if isinstance(a, Result):
      return self.result_compare(a, b)
    assert False

  def relative(self, a, b):
    if isinstance(a, Configuration):
      return self.config_relative(a, b)
    if isinstance(a, Result):
      return self.result_relative(a, b)
    assert None

  def lt(self, a, b):
    return self.compare(a, b) < 0

  def lte(self, a, b):
    return self.compare(a, b) <= 0

  def gt(self, a, b):
    return self.compare(a, b) > 0

  def gte(self, a, b):
    return self.compare(a, b) >= 0

  def min(self, *l):
    if len(l) == 1:
      l = l[0]
    rv = l[0]
    for i in l[1:]:
      if self.lt(i, rv):
        rv = i
    return rv

  def max(self, *l):
    if len(l) == 1:
      l = l[0]
    rv = l[0]
    for i in l[1:]:
      if self.gt(i, rv):
        rv = i
    return rv

  def limit_from_config(self, config):
    """
    a time limit to kill a result after such that it can be compared to config
    """
    results = self.driver.results_query(config=config)
    if results.count() == 0:
      return None
    else:
      return max(list(map(_.time, self.driver.results_query(config=config))))


  def project_compare(self, a1, a2, b1, b2, factor=1.0):
    """
    linearly project both a and b forward to see how they will compare in the
    future
    """
    a3 = Result()
    b3 = Result()
    a3.time = _project(a1.time, a2.time, factor)
    a3.accuracy = _project(a1.accuracy, a2.accuracy, factor)
    a3.energy = _project(a1.energy, a2.energy, factor)
    a3.confidence = _project(a1.confidence, a2.confidence, factor)
    return self.result_compare(a3, b3)

  def display(self, result):
    """
    produce a string version of a resultsdb.models.Result()
    """
    rv = []
    for k in ('time', 'accuracy', 'energy', 'size', 'confidence'):
      v = getattr(result, k)
      if v is not None:
        rv.append('%s=%.4f' % (k, float(v)))
    return ', '.join(rv)

  def filter_acceptable(self, query):
    """Return a Result() query that only returns acceptable results"""
    return query

  def is_acceptable(self, result):
    """Test if a Result() meets thresholds"""
    return True

  def stats_quality_score(self, result, worst_result, best_result):
    """return a score for statistics"""
    if not self.is_acceptable(result):
      return worst_result.time
    else:
      return result.time


def _project(a1, a2, factor):
  if a1 is None or a2 is None:
    return None
  return a2 + factor * (a2 - a1)


class MinimizeTime(SearchObjective):
  """
  minimize Result().time
  """

  def result_order_by_terms(self):
    """return database columns required to order by the objective"""
    return [Result.time]

  def result_compare(self, result1, result2):
    """cmp() compatible comparison of resultsdb.models.Result"""
    return cmp(result1.time, result2.time)

  def config_compare(self, config1, config2):
    """cmp() compatible comparison of resultsdb.models.Configuration"""
    return cmp(min(list(map(_.time, self.driver.results_query(config=config1)))),
               min(list(map(_.time, self.driver.results_query(config=config2)))))

  def result_relative(self, result1, result2):
    """return None, or a relative goodness of resultsdb.models.Result"""
    if result2.time == 0:
      return float('inf') * result1.time
    return old_div(result1.time, result2.time)


class MaximizeAccuracy(SearchObjective):
  """
  maximize Result().accuracy
  """

  def result_order_by_terms(self):
    """return database columns required to order by the objective"""
    return [-Result.accuracy]

  def result_compare(self, result1, result2):
    """cmp() compatible comparison of resultsdb.models.Result"""
    # note opposite order
    return cmp(result2.accuracy, result1.accuracy)

  def result_relative(self, result1, result2):
    """return None, or a relative goodness of resultsdb.models.Result"""
    # note opposite order
    if result1.accuracy == 0:
      return float('inf') * result2.accuracy
    return old_div(result2.accuracy, result1.accuracy)

  def stats_quality_score(self, result, worst_result, best_result):
    """return a score for statistics"""
    if not self.is_acceptable(result):
      return worst_result.time
    else:
      return result.time

  def stats_raw_score(self, result):
    return result.accuracy


class MaximizeAccuracyMinimizeSize(MaximizeAccuracy):
  """
  maximize Result().accuracy, break ties with Result().size
  """

  def result_order_by_terms(self):
    """return database columns required to order by the objective"""
    return [-Result.accuracy, Result.size]

  def result_compare(self, result1, result2):
    """cmp() compatible comparison of resultsdb.models.Result"""
    return cmp((-result1.accuracy, result1.size),
               (-result2.accuracy, result2.size))

  def display(self, result):
    """
    produce a string version of a resultsdb.models.Result()
    """
    return "accuracy=%.8f, size=%.1f" % (result.accuracy, result.size)

  def result_relative(self, result1, result2):
    """return None, or a relative goodness of resultsdb.models.Result"""
    # unimplemented for now
    log.warning('result_relative() not yet implemented for %s',
                self.__class__.__name__)
    return None


class ThresholdAccuracyMinimizeTime(SearchObjective):
  """
  if accuracy >= target:
    minimize time
  else:
    maximize accuracy
  """

  def __init__(self, accuracy_target, low_accuracy_limit_multiplier=10.0):
    self.accuracy_target = accuracy_target
    self.low_accuracy_limit_multiplier = low_accuracy_limit_multiplier
    super(ThresholdAccuracyMinimizeTime, self).__init__()

  def result_order_by_terms(self):
    """return database columns required to order by the objective"""

    return [text("min(accuracy, %f) desc" % self.accuracy_target),
            opentuner.resultsdb.models.Result.time]

  def result_compare(self, result1, result2):
    """cmp() compatible comparison of resultsdb.models.Result"""
    return cmp((-min(self.accuracy_target, result1.accuracy),
                result1.time),
               (-min(self.accuracy_target, result2.accuracy), result2.time))

  def config_compare(self, config1, config2):
    """cmp() compatible comparison of resultsdb.models.Configuration"""
    return self.result_compare(
      self.driver.results_query(config=config1, objective_ordered=True)[0],
      self.driver.results_query(config=config2, objective_ordered=True)[0])

  def limit_from_config(self, config):
    """
    a time limit to kill a result after such that it can be compared to config
    """
    results = self.driver.results_query(config=config)
    if results.count() == 0:
      return None
    if self.accuracy_target > min(list(map(_.accuracy, results))):
      m = self.low_accuracy_limit_multiplier
    else:
      m = 1.0
    return m * max(list(map(_.time, results)))


  def filter_acceptable(self, query):
    """Return a Result() query that only returns acceptable results"""
    return query.filter(opentuner.resultsdb.models.Result.accuracy
                        >= self.accuracy_target)

  def is_acceptable(self, result):
    """Test if a Result() meets thresholds"""
    return result.accuracy >= self.accuracy_target

  def result_relative(self, result1, result2):
    """return None, or a relative goodness of resultsdb.models.Result"""
    # unimplemented for now
    log.warning('result_relative() not yet implemented for %s',
                self.__class__.__name__)
    return None








from __future__ import absolute_import
from builtins import map
from builtins import range
from builtins import object
import abc
import copy
import random
from .technique import SearchTechnique
from opentuner.search import technique

class GlobalEvolutionaryTechnique(SearchTechnique):
  def __init__(self,
               mutation_rate = 0.1,
               crossover_rate = 0.0,
               must_mutate_count = 1,
	             crossover_strength = 0.1,
               *pargs, **kwargs):
    super(GlobalEvolutionaryTechnique, self).__init__(*pargs, **kwargs)
    self.mutation_rate = mutation_rate
    self.crossover_rate = crossover_rate
    self.must_mutate_count = must_mutate_count
    self.crossover_strength = crossover_strength

  @classmethod
  def get_hyper_parameters(cls):
    return ['mutation_rate', 'crossover_rate', 'must_mutate_count', 'crossover_strength']

  def desired_configuration(self):
    """
    return a (cfg, priority) that we should test,
    through random mutation and crossover
    """
    #TODO: set limit value

    parents = self.selection()
    parents = list(map(copy.deepcopy, parents))
    parent_hashes = list(map(self.manipulator.hash_config, parents))

    if len(parents) > 1:
      cfg = self.crossover(parents)
    else:
      cfg = parents[0]

    for z in range(10): #retries
      self.mutation(cfg)
      if self.manipulator.hash_config(cfg) in parent_hashes:
        continue # try again
      return cfg

  def mutation(self, cfg):
    """
    mutate cfg in place
    """
    params = self.manipulator.parameters(cfg)
    random.shuffle(params)
    for param in params[:self.must_mutate_count]:
      self.mutate_param(cfg, param)
    for param in params[self.must_mutate_count:]:
      if random.random() < self.mutation_rate:
        self.mutate_param(cfg, param)

  def mutate_param(self, cfg, param):
    """
    mutate single parameter of cfg in place
    """
    param.op1_randomize(cfg)

  def crossover(self, cfgs):
    cfg1, cfg2, = cfgs
    new = self.manipulator.copy(cfg1)
    params = self.manipulator.parameters(cfg1)
    random.shuffle(params)
    d = int(self.crossover_strength*len(params))
    for param in params[:d]:
      param.set_value(new, param.get_value(cfg2))
    return new

  def selection(self):
    """return a list of parent configurations to use"""
    if random.random() < self.crossover_rate:
      return [self.select(),
              self.select()]
    else:
      return [self.select()]

  @abc.abstractmethod
  def select(self):
    """return a single random parent configuration"""
    return None

class GreedySelectionMixin(object):
  """
  EvolutionaryTechnique mixin for greedily selecting the best known
  configuration
  """
  def select(self):
    """return a single random parent configuration"""
    if (self.driver.best_result is not None and
        self.driver.best_result.state == 'OK'):
      return self.driver.best_result.configuration.data
    else:
      return self.manipulator.random()

class NormalMutationMixin(object):
  """
  Mutate primitive parameters according to normal distribution
  """

  def __init__(self, sigma = 0.1, *pargs, **kwargs):
    super(NormalMutationMixin, self).__init__(*pargs, **kwargs)
    self.sigma = sigma

  def mutate_param(self, cfg, param):
    """
    mutate single parameter of cfg in place
    """
    if param.is_primitive():
      param.op1_normal_mutation(cfg, self.sigma)
    else:
      random.choice(param.manipulators(cfg))(cfg)


class UniformGreedyMutation(GreedySelectionMixin, GlobalEvolutionaryTechnique):
  pass

class NormalGreedyMutation(NormalMutationMixin, GreedySelectionMixin, GlobalEvolutionaryTechnique):
  pass

technique.register(NormalGreedyMutation( crossover_rate=0.5, crossover_strength=0.2, name='GGA'))



from __future__ import print_function
from builtins import range
from builtins import object
import argparse
import copy
import logging
import os
import sys

from datetime import datetime
from fn import _
from opentuner.driverbase import DriverBase
from opentuner.resultsdb.models import Configuration
from opentuner.resultsdb.models import DesiredResult
from opentuner.resultsdb.models import Result
from opentuner.resultsdb.models import BanditInfo
from opentuner.resultsdb.models import BanditSubTechnique
from opentuner.search import plugin
from opentuner.search import technique
from opentuner.search.bandittechniques import AUCBanditMetaTechnique

log = logging.getLogger(__name__)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--test-limit', type=int, default=5000,
                       help='stop tuning after given tests count')
argparser.add_argument('--stop-after', type=float,
                       help='stop tuning after given seconds')
argparser.add_argument('--parallelism', type=int, default=4,
                       help='how many tests to support at once')
argparser.add_argument('--pipelining', type=int, default=0,
                       help='how long a delay (in generations) before results are available')
argparser.add_argument('--bail-threshold', type=int, default=500,
                       help='abort if no requests have been made in X generations')
argparser.add_argument('--no-dups', action='store_true',
                       help='don\'t print out warnings for duplicate requests')
argparser.add_argument('--seed-configuration', action='append', default=[],
                       metavar='FILENAME', help="""
                           Start search at a given configuration.  Can be
                           specified multiple times.  Configurations are loaded
                           with ConfigurationManipulator.load_from_file()
                           and file format is detected from extension.""")


class SearchDriver(DriverBase):
  """
  controls the search process managing root_technique and creating
  DesiredResults
  """

  def __init__(self, manipulator, extra_seeds=None, extra_criteria=None, **kwargs):
    super(SearchDriver, self).__init__(**kwargs)
    if extra_seeds is None:
      extra_seeds = []
    self.manipulator = manipulator
    self.wait_for_results = self.tuning_run_main.results_wait
    self.commit = self.tuning_run_main.commit
    self.extra_criteria = extra_criteria

    self.generation = 0
    self.test_count = 0
    self.plugins = plugin.get_enabled(self.args)
    self.pending_result_callbacks = list()  # (DesiredResult, function) tuples
    # deepcopy is required to have multiple tuning runs in a single process
    if self.args.list_techniques:
      techniques, generators = technique.all_techniques()
      for t in techniques:
        print(t.name)
      sys.exit(0)

    if self.args.generate_bandit_technique:
      # generate a bandit
      self.root_technique = AUCBanditMetaTechnique.generate_technique(manipulator)
    else:
      self.root_technique = copy.deepcopy(technique.get_root(self.args))

    if isinstance(self.root_technique, AUCBanditMetaTechnique):
      self.session.flush()
      info = BanditInfo(tuning_run=self.tuning_run,
                        c=self.root_technique.bandit.C,
                        window=self.root_technique.bandit.window,)
      self.session.add(info)
      for t in self.root_technique.techniques:
        subtechnique = BanditSubTechnique(bandit_info=info,
                                          name=t.name)
        self.session.add(subtechnique)


    self.objective.set_driver(self)
    self.pending_config_ids = set()
    self.best_result = None
    self.new_results = []

    for t in self.plugins:
      t.set_driver(self)
    self.root_technique.set_driver(self)
    self.seed_cfgs = list(extra_seeds)
    for cfg_filename in reversed(self.args.seed_configuration):
      if os.path.exists(cfg_filename):
        self.seed_cfgs.append(manipulator.load_from_file(cfg_filename))
      else:
        log.error('no such file for --seed-configuration %s', cfg_filename)

    self.plugins.sort(key=_.priority)

  def add_plugin(self, p):
    if p in self.plugins:
      return
    self.plugins.append(p)
    self.plugins.sort(key=_.priority)
    p.set_driver(self)

  def convergence_criteria(self):
    """returns true if the tuning process should stop"""
    if self.args.stop_after:
      elapsed = (datetime.now() - self.tuning_run.start_date)
      try:
        elapsed = elapsed.total_seconds()
      except:  # python 2.6
        elapsed = elapsed.days * 86400 + elapsed.seconds
      if elapsed > self.args.stop_after:
          return True
    if self.test_count > self.args.test_limit:
        return True    
    if self.extra_criteria:
        if self.extra_criteria(self.new_results):
            return True
    return False

  def register_result_callback(self, desired_result, callback):
    if desired_result.result is not None:
      callback(desired_result.result)
    else:
      self.pending_result_callbacks.append((desired_result, callback))

  def result_callbacks(self):
    pending = self.pending_result_callbacks
    self.pending_result_callbacks = list()
    for dr, callback in pending:
      if dr.result is not None:
        callback(dr.result)
        continue
      elif self.generation - dr.generation > self.args.pipelining:
        # see if we can find a result
        results = self.results_query(config=dr.configuration).all()
        log.warning("Result callback %d (requestor=%s) pending for "
                    "%d generations %d results available",
                    dr.id, dr.requestor, self.generation - dr.generation,
                    len(results))
        if len(results):
          dr.result = results[0]
          callback(dr.result)
          continue
      # try again later
      self.pending_result_callbacks.append((dr, callback))

  def has_results(self, config):
    return self.results_query(config=config).count() > 0

  def run_generation_techniques(self):
    tests_this_generation = 0
    self.plugin_proxy.before_techniques()
    for z in range(self.args.parallelism):
      if self.seed_cfgs:
        config = self.get_configuration(self.seed_cfgs.pop())
        dr = DesiredResult(configuration=config,
                           requestor='seed',
                           generation=self.generation,
                           request_date=datetime.now(),
                           tuning_run=self.tuning_run)
      else:
        dr = self.root_technique.desired_result()
      if dr is None or dr is False:
        log.debug("no desired result, skipping to testing phase")
        break
      self.session.flush()  # populate configuration_id
      duplicates = (self.session.query(DesiredResult)
                    .filter_by(tuning_run=self.tuning_run,
                               configuration_id=dr.configuration_id)
                    .filter(DesiredResult.id != dr.id)
                    .order_by(DesiredResult.request_date)
                    .limit(1).all())
      self.session.add(dr)
      if len(duplicates):
        if not self.args.no_dups:
          log.warning("duplicate configuration request #%d %s/%s %s",
                      self.test_count,
                      dr.requestor,
                      duplicates[0].requestor,
                      'OLD' if duplicates[0].result else 'PENDING')
        self.session.flush()
        desired_result_id = dr.id

        def callback(result):
          dr = self.session.query(DesiredResult).get(desired_result_id)
          dr.result = result
          dr.state = 'COMPLETE'
          dr.start_date = datetime.now()

        self.register_result_callback(duplicates[0], callback)
      else:
        log.debug("desired result id=%d, cfg=%d", dr.id, dr.configuration_id)
        dr.state = 'REQUESTED'
      self.test_count += 1
      tests_this_generation += 1
    self.plugin_proxy.after_techniques()
    return tests_this_generation

  def process_new_results(self):
    self.new_results = []
    for result in (self.results_query()
                       .filter_by(was_new_best=None)
                       .order_by(Result.collection_date)):
      self.plugin_proxy.on_result(result)
      self.new_results.append(result)
      if self.best_result is None:
        self.best_result = result
        result.was_new_best = True
      elif self.objective.lt(result, self.best_result):
        self.best_result = result
        result.was_new_best = True
        self.plugin_proxy.on_new_best_result(result)
      else:
        result.was_new_best = False
    self.result_callbacks()

  def run_generation_results(self, offset=0):
    self.commit()
    self.plugin_proxy.before_results_wait()
    self.wait_for_results(self.generation + offset)
    self.plugin_proxy.after_results_wait()
    self.process_new_results()

  @property
  def plugin_proxy(self):
    """
    forward any method calls on the returned object to all plugins
    """
    plugins = self.plugins

    class PluginProxy(object):
      def __getattr__(self, method_name):
        def plugin_method_proxy(*args, **kwargs):
          rv = []
          for plugin in plugins:
            rv.append(getattr(plugin, method_name)(*args, **kwargs))
          return [x for x in rv if x is not None]

        return plugin_method_proxy

    return PluginProxy()

  def get_configuration(self, cfg):
    """called by SearchTechniques to create Configuration objects"""
    self.manipulator.normalize(cfg)
    hashv = self.manipulator.hash_config(cfg)
    config = Configuration.get(self.session,self.program, hashv, cfg)
    return config

  def main(self):
    self.plugin_proxy.set_driver(self)
    self.plugin_proxy.before_main()

    no_tests_generations = 0

    # prime pipeline with tests
    for z in range(self.args.pipelining):
      self.run_generation_techniques()
      self.generation += 1

    while not self.convergence_criteria():
      if self.run_generation_techniques() > 0:
        no_tests_generations = 0
      elif no_tests_generations <= self.args.bail_threshold:
        no_tests_generations += 1
      else:
        break
      self.run_generation_results(offset=-self.args.pipelining)
      self.generation += 1

    self.plugin_proxy.after_main()

  def external_main_begin(self):
    self.plugin_proxy.set_driver(self)
    self.plugin_proxy.before_main()

  def external_main_generation(self):
    if self.generation > 0:
      self.plugin_proxy.after_results_wait()
    self.process_new_results()
    self.run_generation_techniques()
    self.commit()
    self.plugin_proxy.before_results_wait()

  def external_main_end(self):
    self.plugin_proxy.after_main()










from __future__ import division
from __future__ import print_function
from builtins import zip
from past.utils import old_div
import argparse
import logging
import time
import socket
import os
from multiprocessing.pool import ThreadPool
from datetime import datetime

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound

from opentuner.driverbase import DriverBase
from opentuner.resultsdb.models import *

log = logging.getLogger(__name__)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--machine-class',
                       help="name of the machine class being run on")


class MeasurementDriver(DriverBase):
  """
  manages the measurement process, reading DesiredResults and creating Results
  """

  def __init__(self,
               measurement_interface,
               input_manager,
               **kwargs):
    super(MeasurementDriver, self).__init__(**kwargs)

    if not self.args.machine_class:
      self.args.machine_class = 'default'

    self.interface = measurement_interface
    self.input_manager = input_manager
    self.commit = self.tuning_run_main.commit
    self.upper_limit_multiplier = 10.0
    self.default_limit_multiplier = 2.0

    self.laptime = time.time()
    self.machine = self.get_machine()

  def get_machine(self):
    """
    get (or create) the machine we are currently running on
    """
    hostname = socket.gethostname()
    try:
      self.session.flush()
      return self.session.query(Machine).filter_by(name=hostname).one()
    except sqlalchemy.orm.exc.NoResultFound:
      m = Machine(name=hostname,
                  cpu=_cputype(),
                  cores=_cpucount(),
                  memory_gb=old_div(_memorysize(), (
                  1024.0 ** 3)) if _memorysize() else 0,
                  machine_class=self.get_machine_class())
      self.session.add(m)
      return m

  def get_machine_class(self):
    """
    get (or create) the machine class we are currently running on
    """
    return MachineClass.get(self.session, name=self.args.machine_class)

  def run_time_limit(self, desired_result, default=3600.0 * 24 * 365 * 10):
    """return a time limit to apply to a test run (in seconds)"""
    best = self.results_query(objective_ordered=True).first()
    if best is None:
      if desired_result.limit:
        return desired_result.limit
      else:
        return default

    if desired_result.limit:
      return min(desired_result.limit, self.upper_limit_multiplier * best.time)
    else:
      return self.default_limit_multiplier * best.time

  def report_result(self, desired_result, result, input=None):
    result.configuration = desired_result.configuration
    result.input = input
    result.machine = self.machine
    result.tuning_run = self.tuning_run
    result.collection_date = datetime.now()
    self.session.add(result)
    desired_result.result = result
    desired_result.state = 'COMPLETE'
    self.input_manager.after_run(desired_result, input)
    result.collection_cost = self.lap_timer()
    self.session.flush()  # populate result.id
    log.debug(
        'Result(id=%d, cfg=%d, time=%.4f, accuracy=%.2f, collection_cost=%.2f)',
        result.id,
        result.configuration.id,
        result.time,
        result.accuracy if result.accuracy is not None else float('NaN'),
        result.collection_cost)
    self.commit()

  def run_desired_result(self, desired_result, compile_result=None,
                         exec_id=None):
    """
    create a new Result using input manager and measurment interface
    Optional compile_result paramater can be passed to run_precompiled as
    the return value of compile()
    Optional exec_id paramater can be passed to run_precompiled in case of
    locating a specific executable
    """
    desired_result.limit = self.run_time_limit(desired_result)

    input = self.input_manager.select_input(desired_result)
    self.session.add(input)
    self.session.flush()

    log.debug('running desired result %s on input %s', desired_result.id,
              input.id)

    self.input_manager.before_run(desired_result, input)

    if self.interface.parallel_compile:
        result = self.interface.run_precompiled(desired_result, input,
                                                desired_result.limit,
                                                compile_result, exec_id)
    else:
        result = self.interface.compile_and_run(desired_result, input,
                                                desired_result.limit)

    self.report_result(desired_result, result, input)

  def lap_timer(self):
    """return the time elapsed since the last call to lap_timer"""
    t = time.time()
    r = t - self.laptime
    self.laptime = t
    return r

  def claim_desired_result(self, desired_result):
    """
    claim a desired result by changing its state to running
    return True if the result was claimed for this process
    """
    self.commit()
    try:
      self.session.refresh(desired_result)
      if desired_result.state == 'REQUESTED':
        desired_result.state = 'RUNNING'
        desired_result.start_date = datetime.now()
        self.commit()
        return True
    except SQLAlchemyError:
      self.session.rollback()
    return False

  def query_pending_desired_results(self):
    q = (self.session.query(DesiredResult)
         .filter_by(tuning_run=self.tuning_run,
                    state='REQUESTED')
         .order_by(DesiredResult.generation,
                   DesiredResult.priority.desc()))
    return q

  def process_all(self):
    """
    process all desired_results in the database
    """
    self.lap_timer()  # reset timer
    q = self.query_pending_desired_results()

    if self.interface.parallel_compile:
      desired_results = []
      thread_args = []

      def compile_result(args):
        interface, data, result_id = args
        return interface.compile(data, result_id)

      for dr in q.all():
        if self.claim_desired_result(dr):
          desired_results.append(dr)
          thread_args.append((self.interface, dr.configuration.data, dr.id))
      if len(desired_results) == 0:
        return
      thread_pool = ThreadPool(len(desired_results))
      # print 'Compiling %d results' % len(thread_args)
      try:
        # Use map_async instead of map because of bug where keyboardinterrupts are ignored
        # See http://stackoverflow.com/questions/1408356/keyboard-interrupts-with-pythons-multiprocessing-pool
        compile_results = thread_pool.map_async(compile_result,
                                                thread_args).get(9999999)
      except Exception:
        # Need to kill other processes because only one thread receives
        # exception
        self.interface.kill_all()
        raise
      # print 'Running %d results' % len(thread_args)
      for dr, compile_result in zip(desired_results, compile_results):
        # Make sure compile was successful
        self.run_desired_result(dr, compile_result, dr.id)
        try:
          self.interface.cleanup(dr.id)
        except RuntimeError as e:
          print(e)
          # print 'Done!'
      thread_pool.close()
    else:
      for dr in q.all():
        if self.claim_desired_result(dr):
          self.run_desired_result(dr)


def _cputype():
  try:
    return re.search(r"model name\s*:\s*([^\n]*)",
                     open("/proc/cpuinfo").read()).group(1)
  except:
    pass
  try:
    # for OS X
    import subprocess

    return subprocess.Popen(["sysctl", "-n", "machdep.cpu.brand_string"],
                            stdout=subprocess.PIPE).communicate()[0].strip()
  except:
    log.warning("failed to get cpu type")
  return "unknown"


def _cpucount():
  try:
    return int(os.sysconf("SC_NPROCESSORS_ONLN"))
  except:
    pass
  try:
    return int(os.sysconf("_SC_NPROCESSORS_ONLN"))
  except:
    pass
  try:
    return int(os.environ["NUMBER_OF_PROCESSORS"])
  except:
    pass
  try:
    return int(os.environ["NUM_PROCESSORS"])
  except:
    log.warning("failed to get the number of processors")
  return 1


def _memorysize():
  try:
    return int(os.sysconf("SC_PHYS_PAGES") * os.sysconf("SC_PAGE_SIZE"))
  except:
    pass
  try:
    return int(os.sysconf("_SC_PHYS_PAGES") * os.sysconf("_SC_PAGE_SIZE"))
  except:
    pass
  try:
    # for OS X
    import subprocess

    return int(subprocess.Popen(["sysctl", "-n", "hw.memsize"],
                                stdout=subprocess.PIPE)
               .communicate()[0].strip())
  except:
    log.warning("failed to get total memory")
  return 1024 ** 3




from past.builtins import cmp
from builtins import zip
from builtins import str
import abc
import logging
from collections import deque, defaultdict
from fn import _

from .technique import SearchTechniqueBase
import sys

log = logging.getLogger(__name__)

class MetaSearchTechnique(SearchTechniqueBase):
  """
  a technique made up of a collection of other techniques
  """
  def __init__(self, techniques, log_freq = 500, *pargs, **kwargs):
    super(MetaSearchTechnique, self).__init__(*pargs, **kwargs)
    self.techniques = techniques
    self.request_count = 0
    self.log_freq = log_freq
    self.logging_use_counters = defaultdict(int)
    self.unique_names()

  def unique_names(self):
    names = set()
    for t in self.techniques:
      while t.name in names:
        t.name += '~'
      t.name = sys.intern(t.name)
      names.add(t.name)

  def set_driver(self, driver):
    super(MetaSearchTechnique, self).set_driver(driver)
    for t in self.techniques:
      t.set_driver(driver)
    self.driver = driver

  def desired_result(self):
    techniques = self.select_technique_order()
    for technique in techniques:
      dr = technique.desired_result()
      if dr is not None:
        if dr is False:
          # technique is waiting for results
          continue
        self.driver.register_result_callback(dr,
            lambda result: self.on_technique_result(technique, result))
        if self.log_freq:
          self.logging_use_counters[technique.name] += 1
          self.debug_log()
        self.request_count += 1
        return dr
      else:
        self.on_technique_no_desired_result(technique)
    return None

  def on_technique_no_desired_result(self, technique):
    """called if a sub-technique returns None"""
    pass

  def on_technique_result(self, technique, result):
    """callback for results of sub-techniques"""
    pass

  @abc.abstractmethod
  def select_technique_order(self):
    """select the order of next techniques to try"""
    return []

  def debug_log(self):
    if self.log_freq and sum(self.logging_use_counters.values())>self.log_freq:
      log.info("%s: %s", self.name,
          str(sorted(list(self.logging_use_counters.items()), key = _[1]*-1)))
      self.logging_use_counters = defaultdict(int)

class RoundRobinMetaSearchTechnique(MetaSearchTechnique):
  """evenly switch between all source techniques"""
  def __init__(self, techniques, **kwargs):
    techniques = deque(techniques)
    super(RoundRobinMetaSearchTechnique, self).__init__(techniques, **kwargs)

  def select_technique_order(self):
    rv = list(self.techniques)
    self.techniques.rotate(1)
    return rv

class RecyclingMetaTechnique(MetaSearchTechnique):
  """
  periodically restart techniques that are not performing well compared to
  global best
  """
  def __init__(self,
               techniques_generators,
               window = 100,
               factor = 5.0,
               **kwargs):
    if 'log_freq' not in kwargs:
      kwargs['log_freq'] = None
    techniques = deque((g(seed_cfg = None) for g in techniques_generators))
    self.rename_i = 0
    for t in techniques:
      self.rename_technique(t)
    super(RecyclingMetaTechnique, self).__init__(techniques, **kwargs)
    self.best_results = defaultdict(lambda: None)
    self.factor = factor
    self.last_check = 0
    self.old_best_results = defaultdict(lambda: None)
    self.technique_generators = deque(techniques_generators)
    self.window = window

  def rename_technique(self, technique):
    technique.name += ".R%d" % self.rename_i
    self.rename_i += 1

  def on_technique_result(self, technique, result):
    """callback for results of sub-techniques"""
    if (self.best_results[technique] is None or
        self.driver.objective.lt(result, self.best_results[technique])):
      self.best_results[technique] = result

  def technique_cmp(self, a, b):
  # a1 = self.old_best_results[a]
  # a2 = self.best_results[a]
  # b1 = self.old_best_results[b]
  # b2 = self.best_results[b]
  # if a1 is None and b1 is None:
  #   return 0
  # if a1 is None:
  #   return -1
  # if b1 is None:
  #   return 1
  # return self.driver.objective.project_compare(a1, a2, b1, b2, self.factor)

    # not ready techniques go to the back
    if not a.is_ready() or not b.is_ready():
      return cmp(b.is_ready(), a.is_ready())

    a = self.best_results[a]
    b = self.best_results[b]
    if a is None and b is None:
      return 0
    if a is None:
      return -1
    if b is None:
      return 1
    return self.driver.objective.compare(a, b)

  def recycle_techniques(self):
    techniques = list(self.techniques)
    techniques.sort(cmp=self.technique_cmp)
    worst = techniques[-1]

    if (not worst.is_ready()
        or (self.old_best_results[worst] is not None
            and self.driver.objective.lt(self.driver.best_result,
                                         self.best_results[worst]))):
      techniques_new = deque()
      tn = None
      for t, gen in zip(self.techniques, self.technique_generators):
        if t is worst:
          tn = gen(seed_cfg=self.driver.best_result.configuration.data)
          self.rename_technique(tn)
          tn.set_driver(self.driver)
          log.info("%s replacing %s with %s", self.name, t.name, tn.name)
          techniques_new.append(tn)
        else:
          techniques_new.append(t)
      self.techniques = techniques_new
    else:
      log.debug("%s: not replacing techniques", self.name)

    self.old_best_results = self.best_results
    self.best_results = defaultdict(lambda: None)
    for t in self.techniques:
      self.best_results[t] = self.old_best_results[t]

  def select_technique_order(self):
    """
    round robin between techniques
    """
    if self.last_check + self.window < self.request_count:
      self.last_check = self.request_count
      self.recycle_techniques()
    rv = list(self.techniques)
    self.techniques.rotate(1)
    self.technique_generators.rotate(1)
    return rv

